# -*- coding: utf-8 -*-
"""a machine consists of 2 parts and has a geometry

"""
from __future__ import print_function
import numpy as np
import logging
from femagtools.dxfsl.shape import Element, Circle, Arc, Line, Shape
from femagtools.dxfsl.corner import Corner
from femagtools.dxfsl.symmetry import Symmetry
import femagtools.dxfsl.area as AREA
from femagtools.dxfsl.functions import point, points_are_close, distance
from femagtools.dxfsl.functions import alpha_angle, normalise_angle, middle_angle, third_angle
from femagtools.dxfsl.functions import alpha_line, line_m, line_n, mirror_point
from femagtools.dxfsl.functions import within_interval, part_of_circle
from femagtools.dxfsl.functions import less, less_equal, greater, greater_equal
from femagtools.dxfsl.journal import Journal, getJournal
logger = logging.getLogger('femagtools.geom')


#############################
#          Machine          #
#############################

class Machine(object):
    def __init__(self, geom, radius=0.0, startangle=0.0, endangle=0.0):
        self.geom = geom
        self.center = self.geom.center
        self.radius = radius
        self.startangle = startangle
        self.endangle = endangle
        self.mirror_orig_geom = None
        self.mirror_geom = None
        self.mirror_startangle = 0.0
        self.mirror_endangle = 0.0
        self.part = self.part_of_circle()
        self.airgaps = []
        self.airgap_radius = 0.0
        self.airgap2_radius = 0.0
        self.airgap_second = None
        self.previous_machine = None
        self.journal = getJournal()

        if not self.center:
            raise ValueError("FATAL ERROR: no center in Geometry")

    def __str__(self):
        return "Machine\n" + \
               "Center: ({})\n".format(self.center) + \
               "Radius: {}\n".format(self.radius) + \
               "Angles: Start={}, End={}\n".format(self.startangle,
                                                   self.endangle) + \
               "Mirror: {}\n".format(self.mirror_geom is not None)

    def log_machine(self, what):
        logger.info("Machine %s", what)
        self.geom.log_geom()

    def is_a_machine(self):
        return self.radius > 0.0

    def is_in_middle(self):
        return self.radius > 0.0 and \
            points_are_close(self.center, [0.0, 0.0], 1e-8)

    def is_full(self):
        return self.radius > 0.0 and \
            self.startangle == 0.0 and self.endangle == 0.0

    def is_half_up(self):
        return self.radius > 0.0 and \
               np.isclose(self.startangle, 0.0, 1e-8) and \
               (np.isclose(self.endangle, np.pi, 1e-8) or
                np.isclose(self.endangle, -np.pi, 1e-8))

    def is_half_down(self):
        return self.radius > 0.0 and \
               (np.isclose(self.endangle, np.pi, 1e-8) or
                np.isclose(self.endangle, -np.pi, 1e-8)) and \
               np.isclose(self.startangle, 0.0, 1e-8)

    def is_half_left(self):
        return self.radius > 0.0 and \
               np.isclose(self.startangle, np.pi/2, 1e-8) and \
               np.isclose(self.endangle, -np.pi/2, 1e-8)

    def is_half_right(self):
        return self.radius > 0.0 and \
               np.isclose(self.startangle, -np.pi/2, 1e-8) and \
               np.isclose(self.endangle, np.pi/2, 1e-8)

    def is_half(self):
        return self.is_half_up() or self.is_half_down() or \
               self.is_half_left() or self.is_half_right()

    def is_quarter(self):
        return self.radius > 0.0 and \
               np.isclose(alpha_angle(self.startangle, self.endangle), np.pi/2)

    def is_startangle_zero(self):
        return np.isclose(self.startangle, 0.0)

    def move_to_middle(self):
        if not self.is_in_middle():
            if self.radius > 0.0:
                offset = [-(self.center[0]), -(self.center[1])]
                self.geom.move(offset)
                self.set_center(0.0, 0.0)
                self.geom.clear_cut_lines()
                return True
        return False

    def set_center(self, x, y):
        self.center[0] = x
        self.center[1] = y
        self.geom.set_center(self.center)

    def set_radius(self, radius):
        self.radius = radius

    def set_kind(self, kind):
        self.geom.kind = kind

    def set_inner_or_outer(self, place=''):
        if place == 'in':
            self.set_inner()
        elif place == 'out':
            self.set_outer()

    def set_inner(self):
        self.geom.is_inner = True
        self.geom.is_outer = False

    def set_outer(self):
        self.geom.is_outer = True
        self.geom.is_inner = False

    def set_attributes(self,
                       kind="",
                       inner=False,
                       outer=False):
        self.set_kind(kind)
        if inner:
            self.set_inner()
        if outer:
            self.set_outer()

    def clear_cut_lines(self):
        self.geom.clear_cut_lines()
        if self.mirror_geom is not None:
            self.mirror_geom.clear_cut_lines()

    def cut_is_possible(self, r_in, r_out):
        if r_in > 0.0:
            if less(self.geom.min_radius, r_in, rtol=0.0001):
                if less_equal(self.geom.max_radius, r_in, rtol=0.0001):
                    return False
                return True
        if r_out > 0.0:
            if greater(self.geom.max_radius, r_out, rtol=0.0001):
                if greater_equal(self.geom.min_radius, r_out, rtol=0.0001):
                    return False
                return True
        return False

    def cut(self, r_in, r_out):
        radius_in = r_in
        radius_out = r_out
        if r_out == 0.0:
            radius_out = self.radius + 10

        clone = self.geom.copy_shape(self.radius,
                                     0.0,
                                     2*np.pi,
                                     radius_in,
                                     radius_out,
                                     False,
                                     append_inner=(r_in > 0.0),
                                     append_outer=(r_out > 0.0))

        if r_out == 0.0:
            r_out = self.radius

        m = Machine(clone,
                    radius=r_out,
                    startangle=self.startangle,
                    endangle=self.endangle)

        m.mirror_geom = self.mirror_geom
        m.part = self.part

        m.set_alfa_and_corners()
        m.set_kind(self.geom.kind)
        return m

    def copy(self, startangle, endangle,
             airgap=False, inside=True, split=False,
             delete_appendices=False,
             concatenate=False,
             connect=False):
        if airgap and self.airgap_radius > 0.0:
            if inside:
                if self.airgap2_radius > 0.0:
                    new_radius = min(self.airgap_radius, self.airgap2_radius)
                else:
                    new_radius = self.airgap_radius
                clone = self.geom.copy_shape(self.radius,
                                             startangle, endangle,
                                             0.0, new_radius,
                                             split=split,
                                             delete_appendices=delete_appendices,
                                             concatenate=concatenate,
                                             connect=connect)
            else:
                new_radius = self.radius
                gap_radius = max(self.airgap_radius, self.airgap2_radius)
                clone = self.geom.copy_shape(self.radius,
                                             startangle, endangle,
                                             gap_radius, self.radius+9999,
                                             split=split,
                                             delete_appendices=delete_appendices,
                                             concatenate=concatenate,
                                             connect=connect)

            circ = Circle(Element(center=self.center,
                                  radius=self.airgap_radius))
            clone.add_cut_line(circ)
        else:
            new_radius = self.radius
            clone = self.geom.copy_shape(self.radius,
                                         startangle, endangle, 0.0,
                                         self.radius+9999,
                                         split=split,
                                         delete_appendices=delete_appendices,
                                         concatenate=concatenate,
                                         connect=connect)

        if not np.isclose(normalise_angle(startangle),
                          normalise_angle(endangle), 0.0):
            start_p = point(self.center, self.radius+5, startangle)
            start_line = Line(Element(start=self.center, end=start_p))
            clone.add_cut_line(start_line)

            end_p = point(self.center, self.radius+5, endangle)
            end_line = Line(Element(start=self.center, end=end_p))
            clone.add_cut_line(end_line)

        if not np.isclose(alpha_angle(startangle, endangle), 2*np.pi):
            return Machine(clone,
                           radius=new_radius,
                           startangle=startangle,
                           endangle=endangle)
        else:
            # Der Originalwinkel bleibt bestehen
            return Machine(clone,
                           radius=new_radius,
                           startangle=self.startangle,
                           endangle=self.endangle)

    def clone(self):
        clone = self.geom.copy_shape(self.radius,
                                     0.0, 2*np.pi,
                                     0.0,
                                     self.radius+9999,
                                     concatenate=False,
                                     connect=False)
        clone.kind = self.geom.kind
        clone.sym_part = self.geom.sym_part
        clone.sym_counterpart = self.geom.sym_counterpart
        clone.alfa = self.geom.alfa
        return Machine(clone,
                       radius = self.radius,
                       startangle=self.startangle,
                       endangle=self.endangle)

    def copy_mirror(self, startangle, midangle, endangle):
        logger.debug("begin of copy_mirror")
        geom1 = self.geom.copy_shape(self.radius,
                                     startangle,
                                     midangle,
                                     0.0,
                                     self.radius+9999,
                                     rtol=1e-08,
                                     atol=1e-08)
        geom2 = self.geom.copy_shape(self.radius,
                                     midangle,
                                     endangle,
                                     0.0,
                                     self.radius+9999,
                                     rtol=1e-08,
                                     atol=1e-08)

        machine = Machine(geom1,
                          radius=self.radius,
                          startangle=startangle,
                          endangle=midangle)
        machine.mirror_orig_geom = self.geom
        machine.mirror_geom = geom2
        machine.mirror_startangle = midangle
        machine.mirror_endangle = endangle
        logger.debug("end of copy_mirror")
        return machine

    def undo_mirror(self):
        assert(self.is_mirrored())
        assert(self.previous_machine)
        self.previous_machine.clear_mirror()
        self.previous_machine.set_alfa_and_corners()
        self.previous_machine.set_kind(self.geom.kind)
        return self.previous_machine

    def clear_mirror(self):
        self.mirror_orig_geom = None
        self.mirror_geom = None
        self.mirror_startangle = 0.0
        self.mirror_endangle = 0.0
        self.geom.mirror_corners = []

    def rotate_to(self, new_startangle):
        if np.isclose(new_startangle, self.startangle):
            return

        if points_are_close(self.center, [0.0, 0.0]):
            angle = new_startangle - self.startangle
            self.geom.area_list = []
            self.geom.rotate(angle)
            self.startangle = new_startangle
            self.endangle += angle

    def airgap(self, correct_airgap=0.0, correct_airgap2=0.0, atol=0.1):
        logger.debug('begin airgap (locking for airgap)')
        self.airgap_radius = 0.0
        self.airgap2_radius = 0.0

        if np.isclose(self.radius, 0.0):
            logger.debug('end airgap: no radius')
            return False

        if correct_airgap < 0:
            logger.debug('end airgap: no airgap (%s)', correct_airgap)
            return False  # no airgap

        self.airgaps = []
        airgaps = self.geom.detect_airgaps(self.center,
                                           self.startangle,
                                           self.endangle, atol)

        alpha = alpha_angle(self.startangle, self.endangle)

        if len(airgaps) == 1:
            self.airgaps = airgaps
        elif len(airgaps) > 0:
            lower_radius = -1.0
            upper_radius = -1.0

            for g in airgaps:
                if np.isclose(g[0], upper_radius):
                    if not self.geom.delete_airgap_circle(self.center,
                                                          lower_radius,
                                                          upper_radius,
                                                          g[1],
                                                          alpha):
                        lower_radius = g[0]
                else:
                    if lower_radius > 0.0:
                        self.airgaps.append((lower_radius, upper_radius))
                    lower_radius = g[0]
                upper_radius = g[1]
            self.airgaps.append((lower_radius, upper_radius))

        if len(self.airgaps) > 0:
            airgap_candidates = []
            prv_radius = 0
            for g in self.airgaps:
                gap_radius = round((g[0]+g[1])/2.0, 6)
                gap_dist = g[1] - g[0]
                prv_dist = g[0] - prv_radius
                prv_radius = g[1]
                circle = Circle(Element(center=self.center,
                                        radius=gap_radius))
                ok, borders = self.geom.is_airgap(self.center,
                                                  self.radius,
                                                  self.startangle,
                                                  self.endangle,
                                                  circle, atol)
                if not ok:
                    logger.error("FATAL: No Airgap with radius {}".
                                 format(gap_radius))
                    self.geom.airgaps.append(circle)
                    return True  # bad exit

                airgap_candidates.append((borders, circle, gap_dist, prv_dist))
                self.geom.airgaps.append(circle)

                if correct_airgap > 0.0:
                    if within_interval(correct_airgap, g[0], g[1], 0.0, 0.0):
                        self.airgap_radius = gap_radius  # ok

                if correct_airgap2 > 0.0:
                    if within_interval(correct_airgap2, g[0], g[1], 0.0, 0.0):
                        self.airgap2_radius = gap_radius  # ok

        if correct_airgap > 0.0 and self.airgap_radius == 0.0:
            logger.error("No airgap with radius {} found"
                         .format(correct_airgap))
            self.show_airgap_candidates(airgap_candidates, False)
            logger.debug("end airgap: bad")
            return True  # bad exit

        if correct_airgap2 > 0.0 and self.airgap2_radius == 0.0:
            logger.error("No airgap2 with radius {} found"
                         .format(correct_airgap2))
            self.show_airgap_candidates(airgap_candidates, False)
            logger.debug("end airgap: bad")
            return True  # bad exit

        if len(self.airgaps) == 0:
            logger.debug('end airgap: No airgap found')
            return False  # no airgaps found

        if self.airgap_radius > 0.0:
            logger.debug("end airgap: radius=%s", self.airgap_radius)
            return False  # correct airgap set

        gaps = [c for b, c, d, prv_d in airgap_candidates if b == 0]

        if len(gaps) == 1:  # one candidate without border intersection
            self.airgap_radius = gaps[0].radius
            logger.debug("end airgap: radius=%s", self.airgap_radius)
            return False  # ok

        if len(airgap_candidates) == 1:  # one candidate found
            self.airgap_radius = airgap_candidates[0][1].radius
            logger.debug("end airgap: radius=%s", self.airgap_radius)
            return False  # ok

        self.airgap_radius = self.show_airgap_candidates(airgap_candidates,
                                                         True)
        logger.debug("end airgap: radius=%s", self.airgap_radius)
        return False  # ok

    def show_airgap_candidates(self, airgap_candidates, get_one):
        if get_one:
            logger.info("{} airgap candidate(s) found:"
                        .format(len(airgap_candidates)))
        else:
            print("{} airgap candidate(s) found:"
                  .format(len(airgap_candidates)))

        dist = 999
        circle = None
        pos_list = []
        for b, c, d, prv_d in airgap_candidates:
            if get_one:
                logger.info(" --- {}   (width={})".format(c.radius, d))
            else:
                print(" --- {}   (width={})".format(c.radius, d))
            if d < dist:
                dist = d
                circle = c

            inner_pc = (c.radius - self.geom.min_radius) / \
                       (self.geom.max_radius - self.geom.min_radius)
            pos = np.abs(inner_pc * 100 - 50)
            logger.debug("Abstand Mitte in % = {}".format(pos))
            prv_dist_percent = int(round(prv_d / self.geom.max_radius, 2) * 100)
            logger.debug("Abstand Vorheriger abs=%s in prz=%s (%s)",
                         prv_d, prv_dist_percent, self.geom.max_radius)
            if pos < 20:
                pos_list.append([pos, d, c])
            elif prv_dist_percent <= 1:
                pos_list.append([prv_dist_percent, d, c])

        if get_one:
            if pos_list:
                dist_list = [[d, c] for pos, d, c in pos_list]
                dist_list.sort()
                circle = dist_list[0][1]
                if len(dist_list) > 1:
                    self.airgap_second = dist_list[1][1]

            logger.info("airgap {} prefered".format(circle.radius))
            return circle.radius

        print("Use options --airgap/--airgap2 <float> to specify")
        return None

    def check_airgap(self):
        return self.geom.check_airgap(self.startangle, self.endangle)

    def install_alternative_airgap(self):
        if self.airgap_second is None:
            return False
        circle = self.airgap_second
        self.airgap_radius = circle.radius
        return True

    def has_alternative_airgap(self):
        return self.airgap_second is not None

    def has_airgap(self):
        return self.airgap_radius > 0.0

    def airgap_x(self):
        return self.airgap_radius

    def airgap_y(self):
        return 0.1

    def part_of_circle(self, pos=2):
        return part_of_circle(self.startangle, self.endangle, pos)

    def delete_center_circle(self):
        gaps = self.geom.get_gaplist(self.center)
        if len(gaps) < 2:
            return

        first_gap = gaps[0][1]
        second_gap = gaps[1][0]
        if first_gap != second_gap:
            return

        first_dist = gaps[0][1]
        second_dist = gaps[1][1] - gaps[1][0]
        if first_dist < 1.0:
            if second_dist / first_dist > 10.0:
                self.geom.delete_circle((0.0, 0.0), first_dist)

    def repair_hull(self):
        logger.debug('begin repair_hull(%s, %s)', self.startangle, self.endangle)
        if self.is_full() and not self.has_airgap():
            self.delete_center_circle()

        if self.startangle == self.endangle:
            logger.debug('end of repair_hull: circle')
            return

        self.repair_hull_geom(self.geom, self.startangle, self.endangle)

        if self.mirror_geom:
            self.repair_hull_geom(self.mirror_geom,
                                  self.mirror_startangle,
                                  self.mirror_endangle)
        logger.debug('end of repair_hull')

    def repair_hull_geom(self, geom, startangle, endangle):
        logger.debug('begin repair_hull_geom (%s, %s)', startangle, endangle)

        rtol = 1e-3
        atol = 1e-3
        c_corner = Corner(self.center, self.center)
        start_c_added, start_corners = geom.get_corner_list(self.center, startangle,
                                                            rtol=rtol, atol=atol)
        end_c_added, end_corners = geom.get_corner_list(self.center, endangle,
                                                        rtol=rtol, atol=atol)
        if start_c_added or end_c_added:
            rtol = 1e-3
            atol = 1e-3
            start_c_added, start_corners = geom.get_corner_list(self.center, startangle,
                                                                rtol=rtol, atol=atol)
            end_c_added, end_corners = geom.get_corner_list(self.center, endangle,
                                                            rtol=rtol, atol=atol)

        geom.repair_hull_line(self.center,
                              startangle, start_corners,
                              c_corner in end_corners,
                              rtol=rtol, atol=atol)
        geom.repair_hull_line(self.center,
                              endangle, end_corners,
                              c_corner in start_corners,
                              rtol=rtol, atol=atol)
        logger.debug('end of repair_hull_geom')

    def create_boundary_nodes(self):
        if self.geom.create_boundary_nodes(self.center, self.startangle, self.endangle):
            logger.debug("___additional boundary nodes created___")

    def create_auxiliary_lines(self):
        logger.debug("create_auxiliary_lines")
        return self.geom.create_auxiliary_lines(self.startangle, self.endangle)

    def set_alfa_and_corners(self):
        self.geom.set_start_corners(self.center, self.startangle)
        self.geom.set_end_corners(self.center, self.endangle)
        self.geom.alfa = alpha_angle(self.startangle, self.endangle)

        if self.mirror_geom is not None:
            self.geom.mirror_corners = self.geom.end_corners
            self.mirror_geom.start_corners = \
                self.mirror_geom.get_corner_nodes(self.center,
                                                  self.mirror_startangle)
            self.mirror_geom.end_corners = \
                self.mirror_geom.get_corner_nodes(self.center,
                                                  self.mirror_endangle)

    def is_mirrored(self):
        return self.mirror_geom is not None

    def num_of_layers(self):
        w = self.geom.num_of_windings()
        if w > 0 and self.geom.winding_is_mirrored():
            return w*2
        return w

    def slot_area(self):
        return self.geom.area_size_of_type(AREA.TYPE_WINDINGS)

    def get_winding_symmetry(self, inside=False):
        logger.debug("begin of get_winding_symmetry")
        symmetry = Symmetry(geom=self.geom,
                            startangle=self.startangle,
                            endangle=self.endangle)
        parts = symmetry.get_winding_symmetry(inside=inside)
        logger.debug("end of get_winding_symmetry (parts=%s)", parts)
        return parts

    def get_magnet_symmetry(self):
        logger.debug("begin of get_magnet_symmetry")
        symmetry = Symmetry(geom=self.geom,
                            startangle=self.startangle,
                            endangle=self.endangle)
        parts = symmetry.get_magnet_symmetry()
        logger.debug("end of get_magnet_symmetry (parts=%s)", parts)
        return parts

    def get_symmetry(self):
        logger.debug("begin of get_symmetry")
        symmetry = Symmetry(geom=self.geom,
                            startangle=self.startangle,
                            endangle=self.endangle)
        parts = symmetry.find_symmetry()
        logger.debug("end of get_symmetry (parts=%s)", parts)
        return parts

    def find_symmetry(self, sym_tolerance, is_inner, is_outer, plt):
        logger.debug("begin of find_symmetry")
        if self.radius <= 0.0:
            return False

        if is_outer:
            self.geom.set_subregion_parameters(self.startangle, self.endangle)
            if self.part > 1:
                self.geom.set_start_corners(self.center, self.startangle, rtol=1e-3, atol=1e-2)
                self.geom.set_end_corners(self.center, self.endangle, rtol=1e-3, atol=1e-2)
            else:
                self.geom.set_virtual_start_end_corners()

            self.geom.looking_for_corners()
            create_areas = self.geom.close_outer_winding_areas()
            if self.geom.adjust_outer_hull_for_symmetry():
                create_areas = True

            if create_areas:
                self.geom.create_list_of_areas(delete=True)

        symmetry = Symmetry(geom=self.geom,
                            startangle=self.startangle,
                            endangle=self.endangle)
        parts = symmetry.find_symmetry()  # temp solution
        logger.debug(">>> Symmetry parts = %s <<<", parts)

        if parts == 1:
            # no slices, but ok
            return False

        found = (parts > 1)

        if not found:
            found = self.geom.find_symmetry(self.radius,
                                            self.startangle,
                                            self.endangle,
                                            sym_tolerance)
        if not found and len(self.geom.area_list) < 5:
            if is_inner:
                found = self.find_stator_symmetry(sym_tolerance, True)
            elif is_outer:
                found = self.find_stator_symmetry(sym_tolerance, False)

        if self.part != 1:  # not full
            logger.debug("end of find_symmetry: not full")
            return found
        return found
    # ENDE
        if found:
            angle = self.geom.sym_slice_angle
            logger.debug(" - #1:  %s slices with angle %s",
                         self.geom.sym_slices,
                         angle)
        else:
            logger.debug(" - #1:  no symmetry found")
            angle = np.pi/2

        if plt:
            plt.render_elements(self.geom, Shape,
                                draw_inside=False,
                                title="Before Copy",
                                show=True,
                                with_corners=False,
                                with_nodes=False,
                                neighbors=True,
                                write_id=True)

        logger.debug("1: ### create a new clone to find symmetry ###")
        elist = self.geom.copy_all_elements(-angle)
        logger.debug(" - %s elements copied", len(elist))
        clone = self.geom.new_clone(elist,
                                    concatenate=False,
                                    connect=False)

        if plt:
            plt.render_elements(clone, Shape,
                                draw_inside=False,
                                title="After Copy",
                                show=True,
                                with_corners=False,
                                with_nodes=False,
                                neighbors=True,
                                write_id=True)

        f = clone.find_symmetry(self.radius,
                                self.startangle,
                                self.endangle,
                                sym_tolerance)
        if f:
            logger.debug(" - #2:  %s slices with angle %s",
                         clone.sym_slices,
                         clone.sym_slice_angle)
        else:
            logger.debug(" - #2:  no symmetry found")

        if not found:
            if f:
                self.geom = clone
            logger.debug("end of find_symmetry: %s slices",
                         self.geom.sym_slices)
            return f

        if f and clone.sym_slices > self.geom.sym_slices:
            self.geom = clone
            logger.debug("end of find_symmetry: %s slices",
                         self.geom.sym_slices)
            return True

        logger.debug("2: ### create a new clone to find symmetry ###")
        elist = clone.copy_all_elements(-angle)
        logger.debug(" - %s elements copied", len(elist))
        clone = clone.new_clone(elist,
                                concatenate=False,
                                connect=False)
        if plt:
            plt.render_elements(clone, Shape,
                                draw_inside=False,
                                title="After Clone",
                                show=True,
                                with_corners=False,
                                with_nodes=False,
                                neighbors=True,
                                write_id=True)

        f = clone.find_symmetry(self.radius,
                                self.startangle,
                                self.endangle,
                                sym_tolerance)
        if f:
            logger.debug(" - #3:  %s slices with angle %s",
                         clone.sym_slices,
                         clone.sym_slice_angle)
        else:
            logger.debug(" - #3:  no symmetry found")

        if f and clone.sym_slices > self.geom.sym_slices:
            self.geom = clone
            logger.debug("end of find_symmetry: %s slices",
                         self.geom.sym_slices)
            return True

        logger.debug("end of find_symmetry: full")
        return found

    def find_stator_symmetry(self, sym_tolerance, is_inner):
        logger.debug("*** Begin of find_stator_symmetry ***")
        if is_inner:
            radius = self.geom.max_radius - 1
        else:
            radius = self.geom.min_radius + 1

        elist = [Circle(Element(center=self.center,
                                radius=radius))]
        elist += self.geom.copy_all_elements(0.0)
        clone = self.geom.new_clone(elist, split=False)
        found = clone.find_symmetry(self.radius,
                                    self.startangle,
                                    self.endangle,
                                    sym_tolerance)
        if found:
            logger.debug(" --> symmetry found <--")
            self.geom.sym_slices = clone.sym_slices
            self.geom.sym_slice_angle = clone.sym_slice_angle
            self.geom.sym_area = clone.sym_area
            self.geom.cut_lines = clone.cut_lines
        logger.debug("*** End of find_stator_symmetry ***")
        return found

    def get_symmetry_slice(self):
        logger.debug("begin get_symmetry_slice")
        if not self.geom.has_symmetry_area():
            logger.debug("end get_symmetry_slice: no symmetry area")
            return None

        machine_slice = self.copy(self.geom.symmetry_startangle(),
                                  self.geom.symmetry_endangle(),
                                  concatenate=True,
                                  connect=True)
        machine_slice.clear_cut_lines()
        machine_slice.repair_hull()
        machine_slice.rotate_to(0.0)
        machine_slice.set_alfa_and_corners()
        logger.debug("end get_symmetry_slice: angle start: {}, end: {}"
                     .format(self.geom.symmetry_startangle(),
                             self.geom.symmetry_endangle()))
        return machine_slice

    def get_forced_winding_slice(self):
        logger.debug("get_forced_winding_slice()")
        if not self.geom.is_outer:
            return None

        areas = len(self.geom.area_list)
        winding_areas = self.geom.num_of_windings()
        iron_areas = self.geom.num_of_irons()
        air_areas = self.geom.num_areas_of_type((AREA.TYPE_AIR,))

        if not (winding_areas + iron_areas + air_areas == areas):
            logger.warning("Warning: strange areas in stator")

        if winding_areas == 0:
            return self.get_possible_windings()

        if winding_areas < 2:
            return None  # nothing to do
        parts = self.get_winding_symmetry()
        if parts < 2:
            return None  # nothing to do
        return self.get_symmetry_slice()

    def get_possible_windings(self, EESM=False, single=False):
        if not self.geom.is_outer:
            return None

        machine = self.clone()
        machine.repair_hull()
        machine.set_alfa_and_corners()
        machine.geom.set_subregion_parameters(self.startangle,
                                              self.endangle)
        machine.geom.looking_for_corners()
        dist_start = machine.geom.dist_start_min_corner()
        dist_end = machine.geom.dist_end_min_corner()

        if not np.isclose(dist_start, dist_end, rtol=1e-3, atol=1e-3):
            angle = machine.geom.alfa
            if dist_start > dist_end:
                machine.mirror_all_areas(self.startangle)
                machine.rotate_to(angle)
                machine.startangle -= angle
            else:
                machine.mirror_all_areas(self.endangle)
                machine.endangle += angle
            machine.geom.create_list_of_areas(delete=True)
            machine.set_alfa_and_corners()
            machine.part = machine.part_of_circle()
            machine.geom.set_subregion_parameters(self.startangle,
                                                  self.endangle)
            machine.geom.looking_for_corners()

        if machine.geom.close_outer_winding_areas():
            machine.geom.create_list_of_areas(delete=True)
            machine.geom.set_subregion_parameters(self.startangle,
                                                  self.endangle)
            machine.geom.looking_for_corners()
            parts = machine.get_winding_symmetry(inside=True)

            if parts == 1:
                return machine
            if parts > 1:
                return machine.get_symmetry_slice()

        return machine

    def get_forced_magnet_slice(self):
        logger.debug("get_forced_magnet_slice()")
        areas = len(self.geom.area_list)
        magnet_areas = self.geom.num_of_magnets()
        iron_areas = self.geom.num_of_irons()
        air_areas = self.geom.num_areas_of_type((AREA.TYPE_AIR,))

        if magnet_areas < 2:
            return None  # nothing to do
        parts = self.get_magnet_symmetry()

        if parts < 2:
            return None  # nothing to do
        slice = self.get_symmetry_slice()
        match = slice.geom.min_max_corners_match()
        while parts > 1 and not match:
            parts = parts - 1
            self.geom.rotate_symmetry_parameters()
            slice = self.get_symmetry_slice()
            match = slice.geom.min_max_corners_match()
        if match:
            slice.geom.set_rotor()
            return slice
        return None

    def get_forced_symmetry(self, part):
        logger.debug("begin get_forced_symmetry")
        if not self.is_full():
            logger.error("it's not a full machine")
            return None

        startangle = 0.0
        endangle = np.pi*2 / part
        machine = self.copy(startangle, endangle)
        machine.clear_cut_lines()
        machine.repair_hull()
        machine.set_alfa_and_corners()
        logger.debug("end get_forced_symmetry")
        return machine

    def get_third_symmetry_mirror(self, plt):
        logger.debug("begin get_third_symmetry_mirror")
        first_thirdangle = third_angle(self.startangle, self.endangle)
        second_thirdangle = middle_angle(first_thirdangle, self.endangle)

        machine_mirror_1 = self.copy_mirror(self.startangle,
                                            first_thirdangle,
                                            second_thirdangle)
        machine_mirror_1.clear_cut_lines()
        machine_mirror_1.repair_hull()
        machine_mirror_1.set_alfa_and_corners()
        if machine_mirror_1.geom.number_of_edges() < 2:  # very bad
            logger.debug("end get_third_symmetry_mirror: no remaining elements")
            return None
        if not machine_mirror_1.check_symmetry_graph(0.001, 0.05):
            logger.debug("end get_third_symmetry_mirror: no mirror first third")
            return None

        if plt:
            plt.render_elements(machine_mirror_1.geom, Shape,
                                title="Part 1",
                                show=True,
                                with_corners=False,
                                with_nodes=False,
                                neighbors=True)

            plt.render_elements(machine_mirror_1.mirror_geom, Shape,
                                title="Counterpart 1",
                                show=True,
                                with_corners=False,
                                with_nodes=False,
                                neighbors=True)

        machine_mirror_2 = self.copy_mirror(first_thirdangle,
                                            second_thirdangle,
                                            self.endangle)
        machine_mirror_2.clear_cut_lines()
        machine_mirror_2.repair_hull()
        machine_mirror_2.set_alfa_and_corners()
        if machine_mirror_2.geom.number_of_edges() < 2:  # very bad
            logger.debug("end get_third_symmetry_mirror: no remaining elements")
            return None
        if not machine_mirror_2.check_symmetry_graph(0.001, 0.05):
            logger.debug("end get_third_symmetry_mirror: no mirror second third")
            return None

        if plt:
            plt.render_elements(machine_mirror_2.geom, Shape,
                                title="Part 2",
                                show=True,
                                with_corners=False,
                                with_nodes=False,
                                neighbors=True)

            plt.render_elements(machine_mirror_2.mirror_geom, Shape,
                                title="Counterpart 2",
                                show=True,
                                with_corners=False,
                                with_nodes=False,
                                neighbors=True)

        machine_mirror_1.previous_machine = self
        logger.debug("end get_third_symmetry_mirror: ok")
        return machine_mirror_1

    def get_symmetry_mirror(self, no_third=False, plt=None):
        logger.debug("begin get_symmetry_mirror")
        if self.part == 1:
            # a complete machine
            startangle = 0.0
            endangle = 0.0
            midangle = np.pi
        else:
            startangle = self.startangle
            endangle = self.endangle
            midangle = middle_angle(self.startangle, self.endangle)
            if no_third:
                machine_mirror = None
            else:
                machine_mirror = self.get_third_symmetry_mirror(plt)
            if machine_mirror:
                logger.debug("end get_symmetry_mirror: third found")
                return machine_mirror

        logger.debug(" - angles: start: {}, mid: {}, end: {}"
                     .format(startangle, midangle, endangle))
        machine_mirror = self.copy_mirror(startangle, midangle, endangle)
        machine_mirror.clear_cut_lines()

        if plt:
            plt.render_elements(machine_mirror.geom, Shape,
                                title="Before Repair Hull",
                                show=True,
                                with_corners=False,
                                with_nodes=False,
                                neighbors=True)

        machine_mirror.repair_hull()
        machine_mirror.set_alfa_and_corners()
        if False:
            plt.render_elements(machine_mirror.geom, Shape,
                                title="Part",
                                show=True,
                                with_corners=False,
                                with_nodes=False,
                                neighbors=True)

            plt.render_elements(machine_mirror.mirror_geom, Shape,
                                title="Counterpart",
                                show=True,
                                with_corners=False,
                                with_nodes=False,
                                neighbors=True)

        if machine_mirror.geom.number_of_edges() < 2:  # very bad
            logger.debug("end get_symmetry_mirror: no remaining elements")
            return None
        if machine_mirror.check_symmetry_graph(0.001, 0.05):
            machine_mirror.previous_machine = self
            machine_mirror.rotate_to(0.0)
            machine_mirror.set_alfa_and_corners()
            logger.debug("end get_symmetry_mirror: found")
            return machine_mirror

        logger.debug("end get_symmetry_mirror: no mirror")
        return None

    def get_symmetry_part(self):
        if self.is_mirrored():
            return self.part/2
        else:
            return self.part

    def get_num_slots(self):
        if self.geom.winding_is_mirrored():
            return self.part/2
        elif self.is_mirrored():
            return self.part/2
        else:
            return self.part

    def get_num_poles(self):
        return self.get_symmetry_part()

    def get_num_parts(self):
        if self.geom.is_rotor():
            return self.get_num_poles()
        if self.geom.is_stator():
            return self.get_num_slots()
        return self.get_symmetry_part()  # strange

    def check_symmetry_graph(self, rtol, atol):
        logger.debug("begin check_symmetry_graph")
        axis_p = point(self.center, self.radius, self.mirror_startangle)
        axis_m = line_m(self.center, axis_p)
        axis_n = line_n(self.center, axis_m)

        def is_node_available(mirror_n, nodes, rtol, atol):
            hits = 0
            for n in nodes:
                if points_are_close(n, mirror_n, rtol, atol):
                    hits += 1
            return hits

        def get_hit_factor(geom1, geom2):
            hit_nodes = 0
            hit_inside = 0
            hit_ag = 0
            hit_no = 0
            if not geom1.g.nodes():
                return 0.0

            def near_opposite_airgap(n, third_radius):
                if not third_radius:
                    return False
                d = distance(self.center, n)
                if geom1.is_inner:
                    return d < third_radius
                else:
                    return d > third_radius

            logger.debug("begin get_hit_factor")

            nodes_near_ag = []
            logger.debug(" -- nodes1: %s,  nodes2: %s",
                         len(geom1.g.nodes()), len(geom2.g.nodes()))
            airgap_radius = geom1.get_airgap_radius()
            opposite_radius = geom1.get_opposite_radius()
            if not (airgap_radius and opposite_radius):
                airgap_radius = self.radius
                opposite_radius = self.radius
                third_radius = None
            else:
                if opposite_radius > airgap_radius:
                    third_radius = airgap_radius + (opposite_radius - airgap_radius) * 0.90
                else:
                    third_radius = airgap_radius - (airgap_radius - opposite_radius) * 0.90
            logger.debug(" -- radius ag: %s,  opposite: %s", airgap_radius, opposite_radius)

            for n in geom1.g.nodes():
                mirror_n = mirror_point(n, self.center, axis_m, axis_n)
                logger.debug("  Node %s <==> %s", n, mirror_n)
                hits = is_node_available(mirror_n, geom2.g.nodes(), rtol=rtol*2, atol=atol*2)
                if hits:
                    logger.debug(" ==> MATCH %s NODE", hits)
                    hit_nodes += 1
                elif near_opposite_airgap(n, third_radius):
                    hits = is_node_available(mirror_n, geom2.g.nodes(), rtol=rtol*10, atol=atol*5)
                    if hits:
                        logger.debug(" ==> MATCH %s NODE (more tolerant)", hits)
                        hit_nodes += 1
                    elif geom2.the_point_is_inside(mirror_n, rtol=rtol*5, atol=atol*2):
                        hit_nodes += 1
                        hits = 1
                if not hits:
                    d = distance(self.center, n)
                    logger.debug(" -- r={}, d={}".format(self.radius, d))
                    if np.isclose(d, airgap_radius, rtol=rtol, atol=atol):
                        hit_ag += 1
                    elif np.isclose(d, airgap_radius, rtol=0.075, atol=atol):
                        # very tolerant
                        nodes_near_ag.append(n)
                    elif geom2.the_point_is_inside(mirror_n, rtol=rtol*10, atol=atol):
                        hit_inside += 1
                    else:
                        hit_no += 1

            if nodes_near_ag:
                logger.debug(" -- %s candidates near airgap --", len(nodes_near_ag))
                for n in nodes_near_ag:
                    alfa = alpha_line(self.center, n)
                    if geom1.is_inner:
                        p = point(self.center, airgap_radius + 10.0, alfa)
                    else:
                        p = point(self.center, airgap_radius - 10.0, alfa)
                    line = Line(Element(start=n, end=p))
                    pts = geom1.intersect_the_line(line)
                    if pts:
                        hit_no += 1
                    else:
                        logger.debug(" -- Node %s is in touch with airgap", n)
                        hit_ag += 1

            logger.debug(" -- temporary nodes hits: %s: factor is %s",
                         hit_nodes, float(hit_nodes) / len(geom1.g.nodes()))
            logger.debug(" -- temporary inside hits: %s: factor is %s",
                         hit_inside, float(hit_inside) / len(geom1.g.nodes()))
            logger.debug(" -- temporary airgap hits: %s: factor is %s",
                         hit_ag, float(hit_ag) / len(geom1.g.nodes()))
            logger.debug(" -- temporary no hits: %s: factor is %s",
                         hit_no, float(hit_no) / len(geom1.g.nodes()))
            # hit_inside and hit_no are unused
            factor = float(hit_nodes + hit_ag) / len(geom1.g.nodes())
            logger.debug("end get_hit_factor => %s", factor)
            return factor

        symmetry = Symmetry(geom=self.geom,
                            startangle=self.startangle,
                            endangle=self.endangle)
        if symmetry.check_symmetry_of_mirror(self.mirror_geom,
                                             self.mirror_startangle):
            logger.debug("NEW SYMMETRY FOUND")
            return True
        return False

        hit_factor1 = get_hit_factor(self.geom,
                                     self.mirror_geom)
        logger.debug("=> hit_factor1 = {}".format(hit_factor1))
        if hit_factor1 < 0.9:
            logger.debug("end check_symmetry_graph: hit_factor1 < 0.9")
            return False  # not ok

        hit_factor2 = get_hit_factor(self.mirror_geom,
                                     self.geom)
        logger.debug("=> hit_factor2 = {}".format(hit_factor2))
        if hit_factor2 < 0.9:
            logger.debug("end check_symmetry_graph: hit_factor2 < 0.9")
            return False  # not ok

        if hit_factor1 < 0.93 and hit_factor2 < 0.93:
            logger.debug("end check_symmetry_graph: hit_factors < 0.93")
            return False  # not ok

        logger.debug("end check_symmetry_graph: ok")
        return True

    def sync_with_counterpart(self, cp_machine):
        logger.debug("sync_with_counterpart")

        def not_tiny_areas(geom):
            sz_list = [a.area_size() for a in geom.list_of_areas()]
            max_sz = max(sz_list)
            large_sz_list = [sz for sz in sz_list if sz > max_sz * 0.005]
            return len(large_sz_list)

        self.geom.sym_counterpart = cp_machine.get_symmetry_part()
        self.geom.sym_part = self.get_symmetry_part()
        logger.debug("part/sym-part: self=%s/%s, cp=%s/%s",
                     self.part, self.geom.sym_part,
                     cp_machine.part, self.geom.sym_counterpart)
        cp_machine.geom.sym_counterpart = self.get_symmetry_part()
        cp_machine.geom.sym_part = cp_machine.get_symmetry_part()

        if not_tiny_areas(self.geom) == 1:
            if not_tiny_areas(cp_machine.geom) > 1:
                self.geom.force_to_be_stator()
                cp_machine.geom.force_to_be_rotor()
        elif not_tiny_areas(cp_machine.geom) == 1:
            self.geom.force_to_be_rotor()
            cp_machine.geom.force_to_be_stator()

    def search_subregions(self, EESM, single=False):
        logger.debug("Search subregions")
        self.geom.search_subregions(self.startangle,
                                    self.endangle,
                                    EESM,
                                    single=single)

    def search_rotor_subregions(self, EESM, single=False):
        if EESM:
            self.geom.search_EESM_rotor_subregions(self.startangle,
                                                   self.endangle,
                                                   single=single)
        else:
            self.geom.search_PMSM_rotor_subregions(self.startangle,
                                                   self.endangle,
                                                   single=single)

    def search_stator_subregions(self, single=False):
        self.geom.search_stator_subregions(self.startangle,
                                           self.endangle,
                                           single=single)

    def rebuild_subregions(self, EESM, single=False):
        logger.debug("Rebuild subregions")
        self.geom.create_list_of_areas(delete=True)
        self.geom.search_subregions(self.startangle,
                                    self.endangle,
                                    EESM,
                                    single=single)

    def has_windings(self):
        return self.geom.has_windings

    def has_magnets(self):
        return self.geom.has_magnets

    def magnets_missing(self, EESM):
        if EESM:
            return False
        return not self.has_magnets()

    def delete_tiny_elements(self, mindist):
        return self.geom.delete_tiny_elements(mindist)

    def search_critical_elements(self, mindist):
        self.geom.search_critical_elements(mindist)

    def create_arc_element(self,
                           radius,
                           startangle,
                           endangle):
        arc = Arc(Element(center=self.center,
                          radius=radius,
                          start_angle=startangle*180/np.pi,
                          end_angle=endangle*180/np.pi),
                  color='darkred',
                  linestyle='dotted')

        node1 = self.geom.find_the_node(arc.p1)
        if node1 is None:
            self.geom.split_with_point(arc.p1)
            node1 = self.geom.find_the_node(arc.p1)
        node2 = self.geom.find_the_node(arc.p2)
        if node2 is None:
            self.geom.split_with_point(arc.p2)
            node2 = self.geom.find_the_node(arc.p2)
        assert(node1 is not None)
        assert(node2 is not None)
        logger.debug("ARC Node1=%s, Node2=%s", node1, node2)
        self.geom.add_element(arc, rtol=1e-3, atol=1e-3)

    def create_arc(self, radius,
                   color='red', linestyle='dotted',
                   attr=None):
        arc = Arc(Element(center=self.center,
                          radius=radius,
                          start_angle=(self.startangle-0.1)*180/np.pi,
                          end_angle=(self.endangle+0.1)*180/np.pi))
        pts = self.geom.split_and_get_intersect_points(arc)
        if len(pts) != 2:
            logger.warning("create_arc(): Bad Points: %s", pts)
            # self.geom.add_edge(arc.node1(4), arc.node2(4), arc)
            return False

        n = self.geom.find_nodes(pts[0], pts[1])
        angle1 = alpha_line(self.center, pts[0])
        angle2 = alpha_line(self.center, pts[1])
        if np.isclose(angle1, self.startangle, atol=1e-03):
            startangle = angle1
            endangle = angle2
        else:
            startangle = angle2
            endangle = angle1
        arc = Arc(Element(center=self.center,
                          radius=radius,
                          start_angle=startangle*180/np.pi,
                          end_angle=endangle*180/np.pi),
                  color=color,
                  linestyle=linestyle)
        arc.set_attribute(attr)

        self.geom.add_or_join_edge(n[0], n[1], arc)
        return True

    def get_iron_separator(self, radius_list):
        if len(radius_list) < 2:
            return 0.0

        r_min = radius_list[0][0]
        for r in radius_list[1:]:
            if np.isclose(r[2], r_min, atol=0.001):
                return r[2]
            r_min = r[0]

        return 0.0

    def has_windings_in_the_middle(self):
        midangle = middle_angle(self.startangle,
                                self.endangle)
        return self.geom.windings_in_the_middle(midangle)

    def has_magnets_in_the_middle(self):
        midangle = middle_angle(self.startangle,
                                self.endangle)
        return self.geom.magnets_in_the_middle(midangle)

    def looking_for_one_possible_magnet(self):
        midangle = middle_angle(self.startangle,
                                self.endangle)
        return self.geom.possible_magnet_in_the_middle(midangle)

    def separate_tooth_and_yoke(self, midangle):
        logger.debug("begin separate_tooth_and_yoke")
        wdg_list = self.geom.get_areas_of_type((AREA.TYPE_WINDINGS,))
        wdg_max_dist = 0.0
        wdg = None
        for w in wdg_list:
            if w.max_dist > wdg_max_dist:
                wdg_max_dist = w.max_dist
                wdg = w

        tooth_list = self.geom.get_areas_of_type((AREA.TYPE_TOOTH,))
        tooth_max_dist = 0.0
        for tooth in tooth_list:
            tooth_max_dist = max(tooth_max_dist, tooth.max_dist)
        tooth_height = tooth_max_dist - self.geom.min_radius
        wdg_height = wdg_max_dist - self.geom.min_radius

        logger.debug("HEIGHT TOOTH=%s, WINDINGS=%s", tooth_height,  wdg_height)

        if greater_equal(tooth_height, wdg_height * 0.9) and \
           less_equal(tooth_height, wdg_height * 1.1):
            logger.debug("Tooths are ok")
            return False  # tooth ok

        iron_list = self.geom.get_areas_of_irons()
        tooth_list = [iron for iron in iron_list
                      if iron.close_to_ag and greater(iron.max_dist, wdg_max_dist, rtol=1e-2, atol=1e-2)]
        logger.debug("max winding dist = %s", wdg_max_dist)
        logger.debug("tooths = %s", len(tooth_list))
        wdg_nodes = [(distance(self.center, n), n) for n in wdg.list_of_nodes()]
        wdg_nodes.sort(reverse=True)

        top_nodes = []
        other_nodes = []
        for d, n in wdg_nodes:
            if np.isclose(d, wdg.max_dist, rtol=1e-2, atol=1e-1):
                top_nodes.append((alpha_line(self.center, n), d, n))
            else:
                other_nodes.append((d, alpha_line(self.center, n), n))

        a_right, d_right, n_right = (None, None, None)
        a_left, d_left, n_left = (None, None, None)

        if not top_nodes:
            other_nodes.sort(reverse=True)
            for d, a, n in other_nodes:
                if a > midangle:
                    if not a_left:
                        a_left = a
                        d_left = d
                        n_left = n
                else:
                    if not a_right:
                        a_right = a
                        d_right = d
                        n_right = n
                if a_left and a_right:
                    break
            if not (a_left and a_right):
                logger.debug("end separate_tooth_and_yoke: no arcs possible")
                return False # bad luck
        else:
            top_nodes.sort()
            a_right, d_right, n_right = top_nodes[0]
            a_left, d_left, n_left = top_nodes[0-1]
        node_right = self.geom.find_the_node(n_right)
        node_left = self.geom.find_the_node(n_left)

        self.create_arc_element(d_right, self.startangle, a_right)
        self.create_arc_element(d_left, a_left, self.endangle)
        logger.debug("end separate_tooth_and_yoke")
        return True

    def create_mirror_lines_outside_windings(self):
        logger.debug("create_mirror_lines_outside_windings")

        if not self.geom.has_areas_touching_both_sides():
            logger.debug("end create_mirror_lines_outside_windings: not done")
            return

        midangle = middle_angle(self.startangle, self.endangle)
        self.separate_tooth_and_yoke(midangle)

        radius = self.radius+10
        ag_list = self.geom.detect_airgaps(self.center,
                                           self.startangle, self.endangle,
                                           atol=0.001,
                                           with_end=True)
        radius_list = [(ag[0], (ag[0] + ag[1]) / 2, ag[1]) for ag in ag_list]
        radius_list.sort(reverse=True)

        line = Line(
            Element(start=self.center,
                    end=point(self.center, radius, midangle)))

        pts = self.geom.split_and_get_intersect_points(line, aktion=False)
        pts.sort()

        arcs_created = False
        p_critical = self.geom.critical_touch_point(pts)
        if p_critical:
            d_critical = distance(self.center, p_critical)
            logger.debug("Critical Point: %s, len=%s", p_critical, d_critical)
            sep_radius = self.get_iron_separator(radius_list)
            logger.debug("Iron Separator found: %s", sep_radius)
            if sep_radius > 0.0 and sep_radius < d_critical:
                radius = sep_radius
            else:
                for r in radius_list:
                    logger.debug("Gap Radius = %s", r[1])
                    if r[1] < d_critical:
                        if self.create_arc(r[1]):
                            radius = r[1]
                            arcs_created = True
                        break

        if arcs_created:
            self.geom.create_list_of_areas(delete=True)
            self.geom.search_subregions(self.startangle,
                                        self.endangle,
                                        False)

        # install line
        line = Line(
            Element(start=self.center,
                    end=point(self.center, radius, midangle)))

        pts = self.geom.split_and_get_intersect_points(line)
        pts.sort()
        if self.geom.create_lines_outside_windings(pts):
            logger.debug("create subregions again")
            self.geom.create_list_of_areas(delete=True)
            self.geom.search_subregions(self.startangle,
                                        self.endangle,
                                        False)

        logger.debug("end create_mirror_lines_outside_windings")

    def create_inner_auxiliary_arcs(self):
        if not self.geom.is_inner:
            return False

        ag_list = self.geom.detect_airgaps(self.center,
                                           self.startangle, self.endangle,
                                           atol=0.001,
                                           with_end=True)
        if not ag_list:
            return False

        done = False
        radius_list = [(ag[0], (ag[0] + ag[1]) / 2, ag[1]) for ag in ag_list]
        min_mag, max_mag = self.geom.get_minmax_magnet()
        radius_list.sort()
        rmin, rmid, rmax = radius_list[0]
        start = 0
        for r in radius_list[1:]:
            if np.isclose(rmax, r[0]):
                # SHAFT
                rmin, rmid, rmax = r
                start = 1
            break

        for r in radius_list[start:]:
            if (r[2] - r[0]) > 4:
                radius = (r[1] + r[2]) / 2
            elif (r[2] - r[1]) > 0.5:
                radius = r[1]
            else:
                radius = min_mag + 1
            if radius < min_mag:
                self.create_arc(radius, attr='iron_sep', color='black')
                done = True
        return done

    def create_mirror_lines_outside_magnets(self):
        logger.debug("create_mirror_lines_outside_magnets")

        rebuild = self.create_inner_auxiliary_arcs()

        if not self.geom.has_areas_touching_both_sides():
            logger.debug("end create_mirror_lines_outside_magnets: not done")
            return rebuild

        radius = self.radius+10
        midangle = middle_angle(self.startangle, self.endangle)
        line = Line(
            Element(start=self.center,
                    end=point(self.center, radius, midangle)))

        pts = self.geom.split_and_get_intersect_points(line)
        pts.sort()

        if self.geom.create_lines_outside_magnets(pts):
            rebuild = True

        logger.debug("end create_mirror_lines_outside_magnets")
        return rebuild

    def create_inner_corner_areas(self):
        self.geom.create_inner_corner_areas(self.startangle, self.endangle)

    def check_and_correct_geom(self, what):
        geom = self.geom.check_geom(what)
        if geom:
            logger.warning("=== Angle correction (%s) ===", what)
            self.geom = geom
            self.startangle = 0.0
            self.endangle = self.geom.alfa
            self.clear_cut_lines()
            self.repair_hull()
            self.set_alfa_and_corners()

    def mirror_all_areas(self, mirror_angle):
        self.geom.mirror_all_areas(mirror_angle)

    def check_airgap_connecting_nodes(self, m_outer):
        logger.info("check_airgap_connecting_nodes")
        assert(self.geom.is_inner)
        assert(m_outer.geom.is_outer)
        self.geom.check_airgap_connecting_nodes(m_outer.geom,
                                                self.startangle,
                                                self.endangle)



# -*- coding: utf-8 -*-
"""
    femagtools.hxy
    ~~~~~~~~~~~~~~

    Reading HXY files (EXPERIMENTAL)
    Build cluster of magnets
"""

import numpy as np
import logging

# K-means clustering
# https://dev.to/sajal2692/coding-k-means-clustering-using-python-and-numpy-fg1
# Sajal Sharma
# with adaptions by Ronald Tanner
def initialize_random_centroids(K, X):
    """Initializes and returns k random centroids"""
    m, n = np.shape(X)
    # a centroid should be of shape (1, n), so the centroids array will be of shape (K, n)
    centroids = np.empty((K, n))
    for i in range(K):
        # pick a random data point from X as the centroid
        centroids[i] = X[np.random.choice(range(m))]
    return centroids


def initialize_centroids(K, X):
    """Initializes and returns k centroids"""
    c = np.mean(X, axis=0)
    if K < 2:
        return [c]
    alfa = np.arctan2(c[1], c[0])  # angle of center axis
    T = np.array([[np.cos(alfa), -np.sin(alfa)],
                  [np.sin(alfa), np.cos(alfa)]])
    p = (np.asarray(X)-c).dot(T) # rotate X
    d = np.linalg.norm(p, axis=1) # distance to center
    a = np.arctan2(p[:, 1], p[:, 0]) # angle to center
    h, b = np.histogram(a, bins=K)
    return [np.mean(X[(a>r[0]) & (a<r[1])], axis=0)
            for r in np.array((b[:-1], b[1:])).T]


def closest_centroid(x, centroids, K):
    """Finds and returns the index of the closest centroid for a given vector x"""
    distances = np.empty(K)
    for i in range(K):
        distances[i] = np.linalg.norm(centroids[i] - x)
    return np.argmin(distances) # return the index of the lowest distance


def create_clusters(centroids, K, X):
    """Returns an array of cluster indices for all the data samples"""
    m, _ = np.shape(X)
    cluster_idx = np.empty(m, dtype=int)
    for i in range(m):
        cluster_idx[i] = closest_centroid(X[i], centroids, K)
    return cluster_idx


def compute_means(cluster_idx, K, X):
    """Computes and returns the new centroids of the clusters"""
    _, n = np.shape(X)
    centroids = np.empty((K, n))
    for i in range(K):
        points = X[cluster_idx == i]  # gather points for the cluster i
        centroids[i] = np.mean(points, axis=0) # use axis=0 to compute means across points
    return centroids


def run_Kmeans(K, X, max_iterations=500):
    """Runs the K-means algorithm and computes the final clusters"""
    # initialize centroids
    centroids = initialize_centroids(K, X)
    # loop till max_iterations or convergance
    logging.debug(f"initial centroids: {centroids}")
    for _ in range(max_iterations):
        # create clusters by assigning the samples to the closet centroids
        clusters = create_clusters(centroids, K, X)
        previous_centroids = centroids
        # compute means of the clusters and assign to centroids
        centroids = compute_means(clusters, K, X)
        # if the new_centroids are the same as the old centroids, return clusters
        diff = previous_centroids - centroids
        if not diff.any():
            return clusters
    return clusters

"""references properties i to magnet k"""
class point():
    def __init__(self, index, k, coord):
        self.index = index
        self.coord = coord
        self.k = k

def readSections(f):
    section = []
    movepos = False
    for line in f:
        if line.startswith(' MOVE POSITION'):
            movepos = True
            if section:
                # skip empty lines
                i = 0
                try:
                    while not section[i]:
                        i = i+1
                except IndexError:
                    i = i-1
                yield section[i:]
                section = []
        if movepos:
            section.append(line.strip())
    yield section


def read(filename, num_magnets):
    """read hxy file and return values grouped to magnets
    returns:
      list of m=num_magnets dicts
        pos: list of n positions in degree
        e: n lists of center coordinates of elements in m
        hxy: n lists of field strengths in kA/m
        bxy: n lists of flux densities in T
        mxy: n lists of magnetization in T
        havg: average of field strengths kA/m
        hmax: maximum of field strengths kA/m
    """
    hxy = []
    with open(filename, encoding='latin1', errors='ignore') as f:
        n = 0
        k = 0
        for s in readSections(f):
            pos = float(s[0].split()[-1])
            num = np.array([[float(x) for x in l.split()] for l in s[5:] if l])
            hxy.append({'pos': pos, 'e': num[:, :2], 'hxy': num[:, 2:4],
                        'bxy': num[:, 4:6], 'mxy':num[:, 6:]})
            logging.info("HXY Section %d: pos %f shape %s", n, pos, num.shape)
            n += 1
            if k == 0:
                k = num.shape[1]
        K = num_magnets
        y_preds = run_Kmeans(num_magnets, np.array(hxy[0]['e']))
        logging.info("Kmeans: %s",
                     [y_preds[y_preds==k].shape[0] for k in range(K)])
        points = [point(i, k, hxy[0]['e'][i]) for i, k in enumerate(y_preds)]
        # move values to magnets:
        magnets = [{'e': [[h['e'][p.index]*1e-3
                           for p in points if p.k == k]
                    for h in hxy]}
                   for k in range(K)]

        for k, mag in enumerate(magnets):
            mag['pos'] = [h['pos'] for h in hxy]
            for mk in ['hxy', 'bxy', 'mxy']:
                mag[mk] = [[h[mk][p.index] for p in points if p.k == k]
                           for h in hxy]
            hxyabs = [np.linalg.norm(hxy, axis=1)
                      for hxy in mag['hxy']]
            mag['havg'] = [np.mean(a) for a in hxyabs]
            mag['hmax'] = [np.max(a) for a in hxyabs]

    return magnets

if __name__ == '__main__':
    import sys
    import matplotlib.pyplot as plt
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')
    magnets = read(sys.argv[1], int(sys.argv[2]))
    for m in magnets:
        print(f"{len(m['e'][0])}: Havg {m['havg']} Hmax {m['hmax']}")

    fig, ax = plt.subplots()
    for m in magnets:
        b = np.array(m['e'][0])
        ax.scatter(*b.T)
    ax.set_aspect('equal')
    ax.autoscale(enable=True)
    ax.axis('off')
    plt.show()



# -*- coding: utf-8 -*-
"""
    femagtools.dxfsl.fslreader
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    renders geometry to fsl


"""
import numpy as np
from .shape import Shape
from .area import TYPE_TOOTH, TYPE_YOKE
import logging
from .. import __version__
logger = logging.getLogger(__name__)


def agndst(da1, da2, Q, p, nodedist=1):
    """ build agndst from set of useful number of nodes
    args:
      da1: stator inner diameter
      da2: rotor outer diameter
      Q: number of stator slots
      p: number of poles
      nodedist: node distance factor
    """
    r = (da1 + da2)/4
    ag = abs(da1 - da2)/6
    lcm = np.lcm(Q, p)//p
    taup = 2*np.pi*r/p
    nmin, nmax = max(1, 18//lcm), 480//lcm
    num_nodes = [i*lcm for i in range(nmin, nmax)
                 if i*lcm % 6 == 0 and i*lcm*p//Q % 2 == 0]
    dagset = np.array([taup/i for i in num_nodes])
    i = np.argmin(np.abs(dagset - ag*nodedist))
    if i > 0 and dagset[i]*nodedist < ag:
        i -= 1
    logger.info("%d Num nodes/p %d Num nodes/slot %g nodedist %g",
                i, num_nodes[i], p*num_nodes[i]/Q, nodedist)
    return dagset[i]


class FslRenderer(object):
    """a model that can created by FSL"""

    def __init__(self, name, mtype):
        self.model = name
        self.mirror_axis = None
        self.fm_nlin = None
        self.shaft = None
        self.agndst = 1
        self.mtype = mtype

    def mirror_nodechains(self, p0, p1):
        self.mirror_axis = np.array((p0, p1)).ravel().tolist()

    def material(self, p0):
        self.fm_nlin = p0

    def circle(self, center, radius,
               color='blue', linestyle=None):
        num = int(2*np.pi*radius)
        if num < 8:
            num = 8
        circle = ['cx, cy = {}, {}'.format(center[0],
                                            center[1]),
                  'nc_circle_m(cx,cy,{}, {})'.format(radius, num),
                  'create_mesh_se(cx, cy)\n']
        self.content += circle

    def arc(self, startangle, endangle, center, radius,
            color='blue', linestyle=None):
        num = 0
        #d = (endangle - startangle) % 2*np.pi
        #num = int(radius*d/self.agndst + 1)
        # if num < 3 and d > np.pi/6:
        #    num = 3
        # else:
        #    num = 0

        p1 = (center[0] + radius*np.cos(startangle),
              center[1] + radius*np.sin(startangle))
        p2 = (center[0] + radius*np.cos(endangle),
              center[1] + radius*np.sin(endangle))
        self.content.append(
            "nc_circle_m({}, {}, {}, {}, {}, {}, {})".format(
                p1[0], p1[1], p2[0], p2[1],
                center[0], center[1], num))

    def ellipse(self, center, width, height,
                rtheta, start_param, end_param,
                color='blue', linestyle=None):
        theta = np.arange(start_param, end_param, 0.2)
        xbegin = 0.5 * width * np.cos(theta)
        ybegin = 0.5 * height * np.sin(theta)
        theta = np.arange(end_param, end_param+0.1, 0.3)
        xend = 0.5 * width * np.cos(theta)
        yend = 0.5 * height * np.sin(theta)

        x = np.concatenate((xbegin, xend), axis=0)
        y = np.concatenate((ybegin, yend), axis=0)

        R = np.array([
            [np.cos(rtheta), -np.sin(rtheta)],
            [np.sin(rtheta),  np.cos(rtheta)]])
        x, y = np.dot(R, np.array([x, y]))
        x += center[0]
        y += center[1]
        nodes = list(zip(x, y))
        n0 = nodes[0]
        for n1 in nodes[1:]:
            self.content.append(
                "nc_line({}, {}, {}, {}, {}) -- ellipse".format(
                    n0[0], n0[1], n1[0], n1[1], 0))
            n0 = n1

    def line(self, p1, p2, e=None,
             color='blue', linestyle=None):
        num = 0
        # if self.nodedist > 0:
        #     l = la.norm(np.asarray(p1)-p2)
        #     num = int(l/self.nodedist + 1)
        if e is not None:
            if e.has_attribute('no_fsl'):
                logger.info("line with attr nofsl")
                return
            if e.has_attribute('auxline'):
                self.content.append(
                    "nc_line({}, {}, {}, {}, {}) -- auxiliary".format(
                        p1[0], p1[1], p2[0], p2[1], num))
                return

        self.content.append(
            "nc_line({}, {}, {}, {}, {})".format(
                p1[0], p1[1], p2[0], p2[1], num))

    def sorted_elements(self, geom, inner=False):
        if inner:
            r = geom.max_radius
        else:
            r = geom.min_radius

        # return sorted list of distances from airgap and elements
        return sorted([(abs(r - np.linalg.norm(e.center_of_connection())), e)
                       for e in geom.elements(Shape)])

    def render(self, machine, inner=False, outer=False, standalone=False):
        '''create fsl statements with nodechains'''
        machine.set_alfa_and_corners()
        geom = machine.geom
        split_len = (geom.max_radius - geom.min_radius) / 4
        geom.split_all_lines_longer_than(split_len)
        self.content = []

        if standalone:
            self.content += ['if (agndst == nil) then',
                             '  agndst = 0.5',
                             '  m.npols_gen = 2',
                             '  m.num_sl_gen = 2',
                             '  new_model_force("{}","Test")'.format(self.model),
                             'end']

        MAXDST=4.0
        NUMLEVELS=10
        NDT0=1.1
        # ndt list format [ (rdistx, ndtx) ...]
        # where
        #   - rdistx is rel dist from airgap (range 0 .. NUMLEVELS-1/NUMLEVELS)
        #   - ndtx nodedist (range NDT0 .. (NUMLEVELS-1/NUMLEVELS))*(MAXDST-1.1) + NDT0)
        ndt_list = [(1.1*nl/NUMLEVELS, nl/NUMLEVELS*(MAXDST-1.1)+NDT0)
                    for nl in range(NUMLEVELS+1)]

        dist = geom.max_radius - geom.min_radius
        el_sorted = self.sorted_elements(geom, inner)

        n = 0
        for d, e in el_sorted:
            d_percent = min(1.0, d / dist)
            if ndt_list[n][0] < d_percent:
                self.agndst = ndt_list[n][1] * self.agndst
                self.content.append('\nndt({}*agndst)\n'.
                                    format(round(ndt_list[n][1], 1)))
                while ndt_list[n][0] < d_percent:
                    n += 1
            e.render(self)

        self.content.append('\n')

        parts = int(machine.get_symmetry_part())
        self.content += ['-- parts      = {}'.format(parts)]
        if geom.is_stator():
            if machine.get_num_slots() > 0:
                self.content += [
                    '-- num_slots  = {}'.format(
                        machine.get_num_slots())
                ]
                # fix stator subregions
                sregs = [a.type for a in geom.list_of_areas()
                         if a.type in (TYPE_YOKE, TYPE_TOOTH)]
                if len(sregs) > 0 and set(sregs) == {TYPE_TOOTH}:
                    for a in geom.list_of_areas():
                        if a.type == TYPE_TOOTH:
                            a.type = TYPE_YOKE
                            logger.debug("FIXED STATOR SUBREGION")

        self.content += ['-- min_radius = {}'.format(geom.min_radius),
                         '-- max_radius = {}'.format(geom.max_radius),
                         '-- min_corner = {}, {}'.format(
                             geom.start_min_corner(0),
                             geom.start_min_corner(1)),
                         '-- max_corner start = {}, {}'.format(
                             geom.start_max_corner(0),
                             geom.start_max_corner(1)),
                         '\n']
        if inner:
            self.content += ['inner_da_start = {}'
                             .format(geom.dist_start_max_corner()),
                             'inner_da_end = {}'
                             .format(geom.dist_end_max_corner())]
        if outer:
            slice = 1 if machine.is_mirrored() else 2
            self.content += [
                '-- create air layer outside',
                'x0, y0 = {}, {}'.format(
                    geom.start_max_corner(0),
                    geom.start_max_corner(1)),
                'hair = 1.0',
                'parts = {}'.format(machine.get_num_parts()),
                'rmax = {}'.format(geom.max_radius),
                'r1 = rmax + hair',
                'r, phi = c2pr(x0, y0)',
                'x1, y1 = pr2c(r1, phi)',
                'x2, y2 = pr2c(r1, {}*math.pi/parts)'.format(slice),
                f'-- end max corner {geom.end_corners[-1]}',
                f'-- center {geom.center}',
                'r2 = {}'.format(geom.dist_end_max_corner(mirrored=False)),
                'x3, y3 = pr2c(r2, {}*math.pi/parts)'.format(slice),
                'nc_line(x0, y0, x1, y1, 0)',
                'nc_circle_m(x1, y1, x2, y2, 0.0, 0.0, 0)',
                'nc_line(x2, y2, x3, y3, 0)',
                'x0, y0 = pr2c(rmax + hair/2, math.pi/parts/2)',
                'create_mesh_se(x0, y0)',
                '\n',
                'outer_da_start = {}'.format(
                    geom.dist_start_min_corner()),
                'outer_da_end = {}'.format(
                    geom.dist_end_min_corner())
            ]
        if self.mtype == 'PMSM':
            self.content += [
                         'xmag = {}',
                         'ymag = {}',
                         'mag_orient = {}']
        self.content += ['\n',
                         'mag_exists = 0',
                         'if mcvkey_yoke == nil then',
                         '  mcvkey_yoke = "dummy"',
                         '  ur = 1000.0',
                         ' else if mcvkey_teeth == nil then',
                         ' mcvkey_teeth = mcvkey_yoke',
                         'end',
                         'end',
                         'x0_iron_tooth, y0_iron_tooth = 0.0, 0.0',
                         'x0_iron_yoke, y0_iron_yoke = 0.0, 0.0',
                         'x0_shaft, y0_shaft = 0.0, 0.0',
                         '\n']

        subregions = {}
        num_windings = 0
        num_magnets = 0
        magor = []

        for area in geom.list_of_areas():
            if area.number_of_elements() > 1:
                p = area.get_point_inside(geom)
                if p and area.type > 0:
                    self.content.append("x0, y0 = {}, {}".format(p[0], p[1]))
                    # self.content.append("point(x0, y0, red, 4)")  # for debugging
                    self.content.append("create_mesh_se(x0, y0)")

                if area.is_winding():
                    if area.type not in subregions:
                        subregions[area.type] = 1
                    if self.mtype == 'PMSM' or outer:
                        num_windings += 1
                        rmin, rmax = area.minmax_dist_from_center((0,0))
                        self.content.append(
                            f'rcoil_{num_windings} = {rmin}, {rmax}')
                        self.content.append('m.xcoil_{}, m.ycoil_{} = x0, y0'.
                                            format(num_windings, num_windings))
                    else:
                        self.content.append('m.xcoil_r, m.ycoil_r = x0, y0')

                elif area.is_magnet():
                    if area.type not in subregions:
                        subregions[area.type] = 1
                    num_magnets += 1
                    self.content.append('xmag[{}], ymag[{}] = x0, y0'.
                                        format(num_magnets, num_magnets))
                    self.content.append('mag_orient[{}] = {}'.
                                        format(num_magnets, area.phi))
                    magor.append([p[0], p[1], area.phi])  # must correct rotation?

                elif area.type > 0:
                    if area.type in subregions:
                        self.content.append(
                            'add_to_subreg(x0, y0, "{}")'.
                            format(area.name()))
                    else:
                        subregions[area.type] = 1
                        color = area.color()
                        if color == 'cyan':
                            color = 'skyblue'
                        self.content.append(
                            'def_new_subreg(x0, y0, "{}", "{}")'.
                            format(area.name(), color))
                    if area.is_stator_iron_yoke():
                        self.content.append(
                            'x0_iron_yoke, y0_iron_yoke = x0, y0')
                    elif area.is_stator_iron_tooth():
                        self.content.append(
                            'x0_iron_tooth, y0_iron_tooth = x0, y0')
                    elif area.is_rotor_iron():
                        self.content.append(
                            'x0_iron_yoke, y0_iron_yoke = x0, y0')
                    elif area.is_shaft():
                        self.content.append(
                            'x0_shaft, y0_shaft = x0, y0')

        self.content.append("create_mesh(x0, y0)\n")

        txt = ["if x0_iron_yoke > 0.0 then",
               "  if mcvkey_yoke ~= 'dummy' then",
               '    def_mat_fm_nlin(x0_iron_yoke, y0_iron_yoke, "blue", mcvkey_yoke, 100)',
               '  else',
               '    def_mat_fm(x0_iron_yoke, y0_iron_yoke, ur, 100)',
               '  end',
               'end\n']
        self.content.append('\n'.join(txt))

        txt = ["if x0_iron_tooth > 0.0 then",
               "  if(x0_iron_yoke == 0 and mcvkey_yoke ~= 'dummy') then",
               "    def_mat_fm_nlin(x0_iron_tooth, y0_iron_tooth, 'blue', mcvkey_yoke, 100)",
               "  else",
               "    if (mcvkey_teeth ~= 'dummy' and mcvkey_teeth ~= nil) then",
               "      def_mat_fm_nlin(x0_iron_tooth, y0_iron_tooth, 'blue', mcvkey_teeth, 100)",
               "    else",
               "      def_mat_fm(x0_iron_tooth, y0_iron_tooth, ur, 100)",
               "    end",
               "  end",
               'end\n']
        self.content.append('\n'.join(txt))

        txt = ['if x0_shaft > 0.0 then',
               "  if mcvkey_shaft ~= 'dummy' then",
               '    def_mat_fm_nlin(x0_shaft, y0_shaft, "lightgrey", mcvkey_shaft, 100)',
               '  else',
               '    def_mat_fm(x0_shaft, y0_shaft, ur, 100)',
               '  end',
               'end\n']
        self.content.append('\n'.join(txt))

        if num_windings > 0:
            if geom.is_mirrored():
                self.content += [
                    'r, phi = c2pr(m.xcoil_1, m.ycoil_1)',
                    'm.xcoil_2, m.ycoil_2 = pr2c(r, {}*2.0 - phi)'.format(geom.alfa)]
            self.content.append('m.wdg_location  = 1 -- stator\n')

        if num_magnets > 0:
            self.content.append('mag_exists   = {}'.
                                format(num_magnets))
            if geom.is_mirrored():
                self.content.append('mag_mirrored = true')
            else:
                self.content.append('mag_mirrored = false')
            self.content.append('mag_alpha    = {}\n'
                                .format(geom.get_alfa()))

        if geom.is_mirrored():
            self.content.append('-- mirror')
            self.content.append('mirror_nodechains({}, {}, {}, {})\n'.format(
                geom.mirror_corners[1][0],   # max x1
                geom.mirror_corners[1][1],   # max y1
                geom.mirror_corners[0][0],   # min x2
                geom.mirror_corners[0][1]))  # min y2

        self.content.append('parts_gen = {}'.format(geom.num_variable()))

        # angle after mirroring
        self.content.append('alfa = {}\n'.format(geom.get_alfa()))

        self.content += ['-- rotate',
                         'x1, y1 = {}, {}'.format(
                             geom.start_corners[0][0],
                             geom.start_corners[0][1])]  # min xy1
        if outer:
            self.content.append('x2, y2 = pr2c(r1, 0.0)')
        else:
            self.content.append('x2, y2 = {}, {}'.format(
                geom.start_corners[1][0],
                geom.start_corners[1][1]))  # max xy1

        if geom.is_mirrored():
            self.content.append('x3, y3 = pr2c(x2, alfa)')
            self.content.append('x4, y4 = pr2c(x1, alfa)')
        elif outer:
            self.content += ['x3, y3 = pr2c(r1, 2*math.pi/parts+phi)',
                             'x4, y4 = {}, {}'.format(
                                 geom.end_corners[0][0],
                                 geom.end_corners[0][1])]  # min xy4
        else:
            self.content += ['x3, y3 = {}, {}'.format(
                                 geom.end_corners[-1][0],
                                 geom.end_corners[-1][1]), # min xy3
                             'x4, y4 = {}, {}'.format(
                                 geom.end_corners[0][0],
                                 geom.end_corners[0][1]),  # min xy4
                             'def_bcond_vpo(x4, y4, x1, y1)']

        self.content.append('if parts_gen > 1 then')
        if geom.corners_dont_match():
            txt = ['  -- Warning: corners dont match',
                   '  mirror_nodechains(x3, y3, x4, y4)',
                   '  if {} > 2 then'.format(geom.num_variable()),
                   '    my_alfa = alfa * 2',
                   '    x3, y3 = pr2c(x2, my_alfa)',
                   '    x4, y4 = pr2c(x1, my_alfa)',
                   '    n = {} / 2'.format(geom.num_variable()),
                   '    rotate_copy_nodechains(x1,y1,x2,y2,x3,y3,x4,y4,n-1)',
                   '  end']
            self.content.append('\n'.join(txt))
        else:
            self.content.append(
                '  rotate_copy_nodechains(x1,y1,x2,y2,x3,y3,x4,y4,parts_gen-1)')
        self.content.append('end')

        self.content.append('alfa = parts_gen * alfa\n')

        if self.fm_nlin:
            self.content.append('\nx0, y0 = {}, {}'. format(
                self.fm_nlin[0], self.fm_nlin[1]))

            mat = ["if fm_nlin_mcvfile ~= 'dummy' then",
                   "  if fm_nlin_mcvfile == 'air' then",
                   "    def_mat_fm(x0,y0, 1.0, fm_nlin_rlen)",
                   "  else",
                   "    def_mat_fm_nlin(x0,y0, fm_nlin_colour, fm_nlin_mcvfile, fm_nlin_rlen)",
                   "  end",
                   "else",
                   "  def_mat_fm(x0,y0, 1000.0, fm_nlin_rlen)",
                   "end"]
            self.content.append('\n'.join(mat))

        if self.shaft:
            mat = ['\nif shaft_mat==1 then',
                   '  def_mat_fm_nlin(0.1,0.1,"lightgrey",fm_nlin_mcvfile_shft,fm_nlin_rlen)',
                   'end']
            self.content.append('\n'.join(mat))

        if num_magnets:
            # TODO: this is clearly a hack to adapt the mag orientation
            # for cases where magnets have multiple small slices
            rotangle = 0
            mag = np.array(magor)
            x, y = mag.T[:2]
            if np.allclose(mag.T[2] - np.arctan2(y, x), np.pi/2, atol=2e-2):
                # all angles are 90°
                rotangle = 90
            self.content += [
                '-- pm magnets',
                'if mag_exists > 0 then',
                '  if( m.remanenc == nil ) then',
                '     m.remanenc = 1.15',
                '     m.relperm  = 1.05',
                '  end',
                '  if( m.orient == nil ) then',
                '    m.orient = m.parallel',
                '  end',
                '  if( m.magncond == nil ) then',
                '    m.magncond = 625000.0',
                '  end',
                '  if( m.rlen == nil ) then',
                '    m.rlen = 100',
                '  end',
                '  for i=0, m.npols_gen-1 do',
                '    for n=1, mag_exists do',
                '      r, p = c2pr(xmag[n], ymag[n])',
                '      x0, y0 = pr2c(r, i*mag_alpha+p)',
                '      phi = (i*mag_alpha+mag_orient[n])*180/math.pi',
                '      if ( i % 2 == 0 ) then',
                f'        phi = phi - {rotangle}',
                '        color = "red"',
                '      else',
                f'        phi = phi - {rotangle} + 180',
                '        color = "green"',
                '      end',
                '      if(m.mcvkey_magnet == nil) then',
                '        def_mat_pm(x0, y0, color, m.remanenc, m.relperm,',
                '                   phi, m.orient, m.magncond, m.rlen)',
                '      else',
                '        def_mat_pm_nlin(x0, y0, color, m.mcvkey_magnet,',
                '                 phi, m.orient, m.magncond, m.rlen)',
                '      end',
                '      if mag_mirrored then',
                '        x0, y0 = pr2c(r, (i+1)*mag_alpha-p)',
                '        phi = ((i+1)*mag_alpha-mag_orient[n])*180/math.pi',
                '        if ( i % 2 == 0 ) then',
                f'          phi = phi - {rotangle}',
                '        else',
                f'          phi = phi - {rotangle} + 180',
                '        end',
                '        if(m.mcvkey_magnet == nil) then',
                '          def_mat_pm(x0, y0, color, m.remanenc, m.relperm,',
                '                       phi, m.orient, m.magncond, m.rlen)',
                '        else',
                '          def_mat_pm_nlin(x0, y0, color, m.mcvkey_magnet,',
                '                   phi, m.orient, m.magncond, m.rlen)',
                '        end',
                '      end',
                '    end',
                '  end',
                'end']

        return self.content

    def render_main(self,
                    m_inner, m_outer,
                    inner, outer,
                    params):
        '''create fsl statements for main file'''
        if not (m_inner and m_outer):
            logger.warning("ERROR: Rotor or Stator missing")
            return []

        if params['external_rotor']:
            stator = inner
            rotor = outer
        else:
            stator = outer
            rotor = inner

        num_layers = min(m_inner.num_of_layers() +
                         m_outer.num_of_layers(),
                         2)
        self.agndst = params.get('agndst', 0.1)
        excwin = []
        if self.mtype == 'EESM':
            excwin = [
                'r,beta  = c2pr(m.xcoil_r, m.ycoil_r)',
                'phi = math.pi/m.num_poles',
                'alpha = phi-beta',
                'm.num_wires = 1',
                'dir = {"wi", "wo"}',
                'xcoil, ycoil = pr2c(r, phi - alpha)',
                'def_new_wdg(xcoil, ycoil, "violet", "Exc", m.num_wires, 10.0, dir[1])',
                'xcoil, ycoil = pr2c(r, phi + alpha)',
                'add_to_wdg(xcoil, ycoil, wsamekey, dir[2], "wser")',
                'for i = 2, m.npols_gen do',
                '  n = (i+1) % 2 + 1',
                '  phi = phi + 2*math.pi/m.num_poles',
                '  xcoil, ycoil = pr2c(r, phi - alpha)',
                '  add_to_wdg(xcoil, ycoil, wsamekey, dir[n], "wser")',
                '  xcoil, ycoil = pr2c(r, phi + alpha)',
                '  n = i % 2 + 1',
                '  add_to_wdg(xcoil, ycoil, wsamekey, dir[n], "wser")',
                'end']
        return [
            '-- generated from DXF by femagtools {}'.format(__version__),
            'exit_on_error = false',
            'exit_on_end = false',
            'verbosity = 2\n',
            'new_model_force("{}","Test")'.format(self.model),
            'global_unit(mm)',
            'pickdist(0.001)',
            'cosys(polar)\n',
            'dy1 = {}'.format(params.get('dy1', 0.0)),
            'da1 = {}'.format(params.get('da1', 0.0)),
            'dy2 = {}'.format(params.get('dy2', 0.0)),
            'da2 = {}'.format(params.get('da2', 0.0)),
            'ag  = (da1 - da2)/2\n',
            'm.tot_num_slot   = {}'.format(params.get('tot_num_slot', 0)),
            'm.num_sl_gen     = {}'.format(params.get('num_sl_gen', 0)),
            'm.num_poles      = {}'.format(params.get('num_poles', 0)),
            'm.num_pol_pair   = m.num_poles/2',
            'm.num_slots      = m.num_sl_gen',
            'm.npols_gen      = m.num_poles * m.num_sl_gen / m.tot_num_slot',
            'm.tot_num_sl     = m.tot_num_slot',
            'm.fc_radius      = (da1+da2)/4',
            'm.fc_radius1     = m.fc_radius',
            'm.arm_length     = 1.0',
            'pre_models("basic_modpar")\n',
            'm.airgap         = 2*ag/3',
            'm.nodedist       = 1.0',
            'agndst           = {}'.format(self.agndst),
            "mcvkey_teeth = 'dummy'",
            "mcvkey_yoke = 'dummy'",
            "mcvkey_shaft = 'dummy'",
            "ur = 1000.0",
            "ndt(agndst)"] + stator + [
                "mcvkey_teeth = 'dummy'",
                "mcvkey_yoke = 'dummy'",
                "mcvkey_shaft = 'dummy'",
                "ndt(agndst)"] + rotor + [
                    '-- airgap',
                    'ndt(agndst)',
                    'r1 = m.fc_radius - ag/6',
                    'x1, y1 = pr2c(r1, alfa)',
                    'n = math.floor(m.fc_radius*alfa/agndst + 1.5)',
                    'nc_circle_m(r1, 0, x1, y1, 0.0, 0.0, n)\n',
                    'r2 = m.fc_radius + ag/6',
                    'x2, y2 = pr2c(r2, alfa)',
                    'nc_circle_m(r2, 0, x2, y2, 0.0, 0.0, n)\n',
                    'if inner_da_start == nil then',
                    '  inner_da_start = da2/2',
                    'end',
                    'x1, y1 = pr2c(inner_da_start, 0.0)',
                    'nc_line(x1, y1, r1, 0.0, 0.0)\n',
                    'if outer_da_start == nil then',
                    '  outer_da_start = da1/2',
                    'end',
                    'x2, y2 = pr2c(outer_da_start, 0.0)',
                    'nc_line(r2, 0.0, x2, y2, 0.0)\n',
                    'if m.tot_num_slot > m.num_sl_gen then',
                    '  if inner_da_end == nil then',
                    '    inner_da_end = inner_da_start',
                    '  end',
                    '  x3, y3 = pr2c(inner_da_end, alfa)',
                    '  x4, y4 = pr2c(r1, alfa)',
                    '  nc_line(x3, y3, x4, y4, 0, 0)\n',
                    '  if outer_da_end == nil then',
                    '    outer_da_end = outer_da_start',
                    '  end',
                    '  x3, y3 = pr2c(outer_da_end, alfa)',
                    '  x4, y4 = pr2c(r2, alfa)',
                    '  nc_line(x3, y3, x4, y4, 0, 0)',
                    'end\n',
                    'x0, y0 = pr2c(r1-ag/6, alfa/2)',
                    'create_mesh_se(x0, y0)',
                    'x0, y0 = pr2c(r2+ag/6, alfa/2)',
                    'create_mesh_se(x0, y0)\n',
                    'connect_models()\n',
                    '-- Gen_winding',
                    'if m.xcoil_1 ~= nil then',
                    '  m.num_phases      = 3',
                    '  m.num_layers      = {}'.format(num_layers),
                    '  m.num_wires       = 1',
                    '  m.coil_span       = {}'.format(
                        params.get(
                            'tot_num_slot', 1)//params.get('num_poles', 1)),
                    '  m.current         = 0.0',
                    '  m.mat_type        = 1.0 -- rotating',
                    '  m.wind_type       = 1.0 -- winding & current',
                    '  m.win_asym        = 1.0 -- sym',
                    '  m.wdg_location    = 1.0 -- stator',
                    '  m.curr_inp        = 0.0 -- const',
                    '  m.dq_offset       = 0',
                    '  pre_models("Gen_winding")',
                    '  pre_models("gen_pocfile")',
                    'end\n',
                    'save_model(cont)\n'] + excwin



"""read a dxf file and create a plot or fsl file

"""
import os
import io
from pathlib import Path
from femagtools import __version__
from femagtools.dxfsl.geom import Geometry
from femagtools.dxfsl.shape import Shape
from femagtools.dxfsl.fslrenderer import FslRenderer, agndst
from femagtools.dxfsl.plotrenderer import PlotRenderer
from femagtools.dxfsl.concat import Concatenation
from femagtools.dxfsl.functions import Timer, middle_angle
from femagtools.dxfsl.journal import Journal, getJournal
from femagtools.dxfsl.area import TYPE_WINDINGS
from femagtools.dxfsl.areabuilder import disable_logging, enable_logging
import logging
import logging.config
import numpy as np
import sys

logger = logging.getLogger(__name__)


def plot_geom(doit, plt, geom, title="Plot", areas=True):
    if not doit or not plt:
        return

    logger.info("Prepare Plot %s", title)
    plt.render_elements(geom, Shape,
                        draw_inside=areas,
                        draw_groups=False,
                        draw_phi=True,
                        critical=True,
                        title=title,
                        show=True,
                        with_corners=True,
                        with_nodes=False,
                        neighbors=True,
                        write_id=areas,
                        with_legend=False,
                        fill_areas=areas)


def symmetry_search(machine,
                    plt=None,  # plotter
                    kind="single",
                    mindist=0.01,
                    symtol=0.0,
                    sympart=0,
                    is_inner=False,
                    is_outer=False,
                    show_plots=True,
                    debug_mode=False,
                    rows=1,
                    cols=1,
                    num=1):
    logger.info("*** Begin symmetry search for %s ***", kind)

    def return_machine(machine, kind):
        machine.set_attributes(kind=kind, inner=is_inner, outer=is_outer)
        machine.check_and_correct_geom(kind)
        machine.delete_tiny_elements(mindist)
        return machine

    machine.set_attributes(kind=kind, inner=is_inner, outer=is_outer)
    machine.clear_cut_lines()
    if show_plots and debug_mode:
        plt.render_elements(machine.geom, Shape,
                            neighbors=True, title=kind)

    if sympart > 0:
        if not machine.is_full():
            logger.error("force symmetry failed")
            sys.exit(1)
        machine_ok = machine.get_forced_symmetry(sympart)
        logger.info("*** End of symmetry search for %s ***", kind)
        return return_machine(machine_ok, kind)

    plot_geom(False,  # for developer
              plt, machine.geom,
              title="Before Symmetry ({})".format(kind))

    if machine.find_symmetry(symtol, is_inner, is_outer, None):
        logger.info(" - {}: symmetry axis found".format(kind))
        plot_geom(False,  # for developer
                  plt, machine.geom,
                  title="Symmetry found")

        if show_plots:
            plt.render_elements(machine.geom, Shape,
                                title=kind+' (symmetrylines)',
                                draw_inside=True,
                                rows=rows, cols=cols, num=num, show=False)
        machine_slice = machine.get_symmetry_slice()
        if machine_slice is None:
            logger.info(" - no slice extracted ?!?")
            return return_machine(machine, kind)

        plot_geom(False,  # for developer
                  plt, machine_slice.geom,
                  title="Slice of {}".format(kind))

        # no third after slice
        machine_mirror = machine_slice.get_symmetry_mirror(no_third=True)

        if machine_mirror is not None:
            logger.info("*** End of symmetry search for %s (symmetry and mirror) ***", kind)
            return return_machine(machine_mirror, kind)

        logger.info("*** End of symmetry search for %s (symmetry) ***", kind)
        return return_machine(machine_slice, kind)

    # --- no symmetry slice found ---
    logger.info(" - {}: no symmetry axis found".format(kind))
    if show_plots:
        if debug_mode:
            plt.render_elements(machine.geom, Shape,
                                title=kind+' (no symmetry axis)',
                                draw_inside=True,
                                neighbors=True,
                                rows=rows, cols=cols, num=num, show=False)
        else:
            plt.add_emptyplot(rows, cols, num, 'no symmetry axis')

    if not machine.is_startangle_zero():
        logger.debug("Rotate geometry to 0.0")
        machine.rotate_to(0.0)
        machine.set_alfa_and_corners()

    plot_geom(False,  # for developer
              plt, machine.geom,
              title="Before Mirror ({})".format(kind))

    machine_mirror = machine.get_symmetry_mirror()
    machine_slice = machine
    machine_slice.set_alfa_and_corners()

    if machine_mirror is None:
        logger.info(" - no mirror found")
        machine_ok = machine_slice
    else:
        plot_geom(False,  # for developer
                  plt, machine_mirror.geom,
                  title="After Mirror ({})".format(kind))

        logger.info(" - mirror found")
        machine_next_mirror = machine_mirror.get_symmetry_mirror()
        while machine_next_mirror is not None:
            logger.info(" - another mirror found")
            machine_mirror = machine_next_mirror
            plot_geom(False,  # for developer
                      plt, machine_mirror.geom,
                      title="Next Mirror ({})".format(kind))
            machine_next_mirror = machine_mirror.get_symmetry_mirror()

        machine_ok = machine_mirror

    if not machine_ok.is_startangle_zero():
        logger.debug("Rotate geometry to 0.0")
        machine_ok.rotate_to(0.0)
        machine_ok.set_alfa_and_corners()

    logger.info("*** End of symmetry search for %s ***", kind)
    return return_machine(machine_ok, kind)


def build_machine_rotor(machine, inner, mindist, plt, EESM=False, single=False):
    logger.debug("Begin of build_machine_rotor")

    if machine.part == 1:
        logger.error("Full machine rotor without symmetrylines")
        sys.exit(1)

    if machine.has_windings():
        logger.debug("do nothing here with windings in rotor")
        logger.debug("End of build_machine_rotor")
        return machine

    timer = Timer(start_it=True)

    if machine.is_mirrored():
        logger.debug("Rotor is mirrored")
        machine_temp = machine.undo_mirror()
        machine_temp.delete_tiny_elements(mindist)
        machine_temp.geom.set_rotor()
        machine_temp.search_rotor_subregions(EESM, single=single)
    else:
        machine_temp = machine

    plot_geom(False,  # for developer
              plt, machine_temp.geom,
              title="Inner Rotor check magnets")

    machine_slice = machine_temp.get_forced_magnet_slice()
    if machine_slice:
        plot_geom(False,  # for developer
                  plt, machine_slice.geom,
                  title="Rotor Magnet Slice")

        machine_temp = machine_slice
        machine_temp.geom.set_rotor()
        machine_temp.rebuild_subregions(EESM, single=single)
        plot_geom(False,  # for developer
                  plt, machine_temp.geom,
                  title="Rotor Magnet Slice after Rebuild")

    rebuild = False
    if machine_temp.magnets_missing(EESM):
        mag = machine_temp.looking_for_one_possible_magnet()
        if mag:
            if machine.is_mirrored():
                rebuild = machine_temp.create_mirror_lines_outside_magnets()
                if machine_temp.create_auxiliary_lines():
                    rebuild = True
                if rebuild:
                    machine_temp.rebuild_subregions(EESM, single=single)
                    machine_temp.geom.force_area_as_magnet(mag)
            return machine_temp

    if machine_temp.has_magnets_in_the_middle():
        logger.debug("Magnets cut")
        rebuild = machine_temp.create_mirror_lines_outside_magnets()
    else:
        if machine.is_mirrored():
            logger.debug("Back to the mirrored machine")
            machine_temp = machine  # undo
            rebuild = machine_temp.create_inner_auxiliary_arcs()
        else:
            rebuild = machine_temp.create_mirror_lines_outside_magnets()
    if rebuild:
        machine_temp.geom.create_list_of_areas(delete=True)

    if machine_temp.create_auxiliary_lines():
        logger.debug("Auxiliary Lines created: rebuild subregions")
        rebuild = True
    if rebuild:
        machine_temp.rebuild_subregions(EESM, single=single)

    machine_temp.geom.recalculate_magnet_orientation()
    if inner:
        machine_temp.create_inner_corner_areas()

    if not machine_temp.is_mirrored():
        plot_geom(False,  # for developer
                  plt, machine_temp.geom,
                  title="Rotor before Boundery Corr")
        machine_temp.create_boundary_nodes()

    plot_geom(False,  # for developer
              plt, machine_temp.geom,
              title="Final Rotor")

    t = timer.stop("-- rotor created in %0.4f seconds --")
    journal.put('time_rotor_created', t)

    logger.debug("End of build_machine_rotor")
    return machine_temp


def build_machine_stator(machine, inner, mindist, plt, EESM=False, single=False):
    logger.debug("Begin of build_machine_stator")
    timer = Timer(start_it=True)

    if machine.part == 1:
        logger.error("Full stator without symmetrylines")
        sys.exit(1)

    if not machine.geom.is_stator():
        logger.debug("Rotor with windings")

    if machine.is_mirrored():
        plot_geom(False,  # for developer
                  plt, machine.previous_machine.geom,
                  title="Mirrored Stator")

        logger.debug("undo mirrored windings")
        machine_temp = machine.undo_mirror()
        machine_temp.delete_tiny_elements(mindist)
        machine_temp.geom.set_stator()
        machine_temp.search_stator_subregions(single=single)
        if not machine_temp.has_windings_in_the_middle():
            logger.debug("Back to the mirrored machine")
            machine_temp = machine  # undo
        else:
            machine_temp.create_mirror_lines_outside_windings()
    else:
        machine_temp = machine
        if machine_temp.has_windings_in_the_middle():
            machine_temp.create_mirror_lines_outside_windings()

    if machine_temp.geom.reduce_element_nodes(mindist):
        machine_temp.rebuild_subregions(EESM, single=single)
        plot_geom(False,  # for developer
                  plt, machine_temp.geom,
                  title="Nodes reduced")

    machine_slice = machine_temp.get_forced_winding_slice()
    if machine_slice:
        plot_geom(False,  # for developer
                  plt, machine_slice.geom,
                  title="Stator Winding Slice")

        machine_temp = machine_slice
        machine_temp.geom.set_stator()
        machine_temp.rebuild_subregions(EESM, single=single)
        if machine_temp.has_windings_in_the_middle():
            machine_temp.create_mirror_lines_outside_windings()

        plot_geom(False,  # for developer
                  plt, machine_temp.geom,
                  title="Stator Winding Slice after Rebuild")

    if machine_temp.create_auxiliary_lines():
        machine_temp.rebuild_subregions(EESM, single=single)
        plot_geom(False,  # for developer
                  plt, machine_temp.geom,
                  title="Stator with Auxiliary Lines")

    if inner:
        machine_temp.create_inner_corner_areas()

    if not machine_temp.is_mirrored():
        plot_geom(False,  # for developer
                  plt, machine_temp.geom,
                  title="Stator before Boundary Corr")
        machine_temp.create_boundary_nodes()

    plot_geom(False,  # for developer
              plt, machine_temp.geom,
              title="Final Stator")

    t = timer.stop("-- stator created in %0.4f seconds --")
    journal.put('time_stator_created', t)

    logger.debug("End of build_machine_stator")
    return machine_temp


def build_inner_machine(machine,
                        mindist=0.01,
                        plt=None,
                        EESM=False):
    logger.info("Begin of build_inner_machine")
    machine.search_subregions(EESM)

    if machine.geom.is_rotor():  # Inner mirrored rotor
        machine = build_machine_rotor(machine,
                                      True,  # is inner
                                      mindist,
                                      plt,
                                      EESM=EESM)

    if machine.geom.is_stator() or machine.has_windings():
        machine = build_machine_stator(machine,
                                       True,
                                       mindist,
                                       plt,
                                       EESM=EESM)

    machine.search_critical_elements(mindist)
    logger.info("End of build_inner_machine")
    return machine


def build_outer_machine(machine,
                        mindist=0.01,
                        plt=None,
                        EESM=False):
    logger.info("Begin of build_outer_machine")
    machine.search_subregions(EESM)

    if machine.geom.is_rotor():  # Outer mirrored rotor
        machine = build_machine_rotor(machine,
                                      False,  # is outer
                                      mindist,
                                      plt,
                                      EESM=EESM)

    if machine.geom.is_stator() or machine.has_windings():
        machine = build_machine_stator(machine,
                                       False,
                                       mindist,
                                       plt,
                                       EESM=EESM)

    machine.search_critical_elements(mindist)
    logger.info("End of build_outer_machine")
    return machine


def convert(dxfile,
            EESM=False,
            rtol=1e-04,
            atol=1e-03,
            mindist=0.01,
            symtol=0.001,
            sympart=0,
            split=False,
            inner_name='inner',
            outer_name='outer',
            part=(),
            airgap=0.0,
            airgap2=0.0,
            da=0.0,
            dy=0.0,
            nodedist=1,
            view_only=False,
            view_korr=False,
            show_plots=False,
            show_areas=False,
            small_plots=False,
            write_fsl=True,
            write_fsl_single=False,
            write_png=False,
            write_id=False,
            full_model=False,
            debug_mode=False,
            write_journal=False):
    global journal
    layers = ()
    conv = {}

    input_file = Path(dxfile)
    if not input_file.is_file():
        logger.error("File %s is not available", input_file)
        sys.exit(1)

    basename = input_file.stem
    if part:
        logger.info("***** start processing %s (%s) [%s] *****",
                    basename,
                    part,
                    __version__)
    else:
        logger.info("***** start processing %s [%s] *****",
                    basename,
                    __version__)
    timer = Timer(start_it=True)
    start_timer = Timer(start_it=True)

    journal = getJournal(name='converter_journal', aktiv=write_journal)
    journal.get_journal(input_file.name)
    journal.set_filename(str(input_file.resolve()))
    journal.set('success', False)
    journal.write_journal()

    if part:
        if part[0] not in ('rotor', 'stator'):
            logger.error('FATAL: Parameter rotor or stator expected')
            return dict(error='unknown part {}'.format(part))
        if part[1] not in ('in', 'out'):
            logger.error('"{}" has to be defined in/out'.format(part[0]))
            return dict(error='unknown location {}'.format(part[1]))
    else:
        if da:
            logger.warning("distance airgap (da) ignored")
            da = 0.0
        if dy:
            logger.warning("distance yoke (dy) ignored")
            dy = 0.0

    split_ini = split
    split_cpy = False
    if not (part or view_only):
        split_ini = False
        split_cpy = split

    if write_fsl_single:
        write_fsl = True
    if small_plots:
        show_plots = False

    try:
        if input_file.suffix in ['.fem', '.FEM']:
            from .femparser import femshapes
            basegeom = Geometry(femshapes(dxfile),
                                rtol=rtol,
                                atol=atol,
                                split=split_ini,
                                concatenate=True,
                                connect=True,
                                delete=True,
                                adjust=True,
                                main=True)
        elif input_file.suffix in ['.dxf', '.DXF']:
            from .dxfparser import dxfshapes
            basegeom = Geometry(dxfshapes(dxfile,
                                          mindist=mindist,
                                          layers=layers),
                                rtol=rtol,
                                atol=atol,
                                split=split_ini,
                                concatenate=True,
                                connect=True,
                                delete=True,
                                adjust=True,
                                main=True)
        elif input_file.suffix in ['.svg', '.SVG']:
            from .svgparser import svgshapes
            basegeom = Geometry(svgshapes(dxfile),
                                rtol=rtol,
                                atol=atol,
                                split=split_ini,
                                concatenate=True,
                                connect=True,
                                delete=True,
                                adjust=True,
                                main=True)
        else:
            logger.error("Unexpected file %s", input_file)
            sys.exit(1)
    except FileNotFoundError as ex:
        logger.error(ex)
        return dict()

    logger.info("total elements %s", len(basegeom.g.edges()))

    p = PlotRenderer()

    if view_only:
        logger.info("View only")
        p.render_elements(basegeom, Shape,
                          neighbors=True,
                          png=write_png,
                          show=True)
        return dict()

    plot_geom(False,  # for developer
              p, basegeom,
              title="Before finding Machine")

    if show_plots:
        p.render_elements(basegeom, Shape,
                          title=input_file.name,
                          with_hull=False,
                          rows=3, cols=2, num=1, show=debug_mode)

    machine_base = basegeom.get_machine()
    if not machine_base.is_a_machine():
        logger.warn("it's Not a Machine")
        return dict(error='machine not detected')

    if not (machine_base.part > 0):
        # machine shape is unclear
        logger.warn("machine shape is unclear")
        machine_base.set_center(0.0, 0.0)
        machine_base.set_radius(9999999)

    machine = machine_base

    if machine.part_of_circle() == 0:
        logger.warn("No arc segment found")
        return dict(error='no arc segment found')

    machine.clear_cut_lines()
    machine.move_to_middle()
    if show_plots and debug_mode:
        p.render_elements(machine.geom, Shape,
                          title='Areas',
                          neighbors=True,
                          with_corners=False, show=True)

    if machine.airgap(airgap, airgap2, symtol):
        p.render_elements(machine.geom, Shape,
                          title='Search for airgap failed',
                          with_corners=False, show=True)
        return dict(error='no airgap found')

    if show_plots:
        p.render_elements(basegeom, Shape, neighbors=True,
                          title='Original with nodes',
                          rows=3, cols=2, num=2, show=False)

    machine.repair_hull()

    if machine.has_airgap():
        logger.info("=== airgap is %s ===", machine.airgap_radius)
        machine_inner = machine.copy(startangle=0.0,
                                     endangle=2*np.pi,
                                     airgap=True,
                                     inside=True,
                                     split=split_cpy,
                                     delete_appendices=True,
                                     concatenate=True,
                                     connect=True)
        machine_inner.set_inner()
        machine_outer = machine.copy(startangle=0.0,
                                     endangle=2*np.pi,
                                     airgap=True,
                                     inside=False,
                                     split=split_cpy,
                                     delete_appendices=True,
                                     concatenate=True,
                                     connect=True)
        machine_outer.set_outer()

        # check airgap
        if machine.has_alternative_airgap():
            bad_inner = machine_inner.check_airgap()
            bad_outer = False
            if not bad_inner:
                bad_outer = machine_outer.check_airgap()
            if bad_inner or bad_outer:
                logger.warning("***** bad Airgap *****")
                if bad_inner:
                    plot_geom(False,  # for developer
                              p, machine_inner.geom,
                              title="Bad Inner Airgap Geometry")
                if bad_outer:
                    plot_geom(False,  # for developer
                              p, machine_outer.geom,
                              title="Bad Outer Airgap Geometry")

                if machine.install_alternative_airgap():
                    logger.info("=== alternative airgap is %s ===",
                                machine.airgap_radius)
                    machine_inner = machine.copy(startangle=0.0,
                                                 endangle=2*np.pi,
                                                 airgap=True,
                                                 inside=True,
                                                 split=split_cpy,
                                                 delete_appendices=True,
                                                 concatenate=True,
                                                 connect=True)
                    machine_inner.set_inner()
                    machine_outer = machine.copy(startangle=0.0,
                                                 endangle=2*np.pi,
                                                 airgap=True,
                                                 inside=False,
                                                 split=split_cpy,
                                                 delete_appendices=True,
                                                 concatenate=True,
                                                 connect=True)
                    machine_outer.set_outer()

        start_timer.stop("-- first part in %0.4f seconds --", info=True)

        process_timer = Timer(start_it=True)
        # inner part
        machine_inner = symmetry_search(machine=machine_inner,
                                        plt=p,  # plot
                                        kind=inner_name,
                                        is_inner=True,
                                        mindist=mindist,
                                        symtol=symtol,
                                        show_plots=show_plots,
                                        rows=3,  # rows
                                        cols=2,  # columns
                                        num=3)   # start num

        # outer part
        machine_outer = symmetry_search(machine_outer,
                                        plt=p,  # plot
                                        kind=outer_name,
                                        is_outer=True,
                                        mindist=mindist,
                                        symtol=symtol,
                                        show_plots=show_plots,
                                        rows=3,  # rows
                                        cols=2,  # columns
                                        num=4)   # start num

        process_timer.stop("-- symmetry search in %0.4f seconds --", info=True)

        machine_inner.sync_with_counterpart(machine_outer)

        final_timer = Timer(start_it=True)
        machine_inner = build_inner_machine(machine_inner,
                                            mindist=mindist,
                                            plt=p,
                                            EESM=EESM)

        machine_outer = build_outer_machine(machine_outer,
                                            mindist,
                                            p,
                                            EESM=EESM)
        final_timer.stop("-- final part in %0.4f seconds --", info=True)

        machine_inner.sync_with_counterpart(machine_outer)
        logger.info("***** END of work: %s *****", basename)

        if machine_inner.geom.is_rotor():
            inner_title = "Rotor"
            outer_title = "Stator"
        else:
            inner_title = "Stator"
            outer_title = "Rotor"

        plot_geom(False,  # for developer
                  p, machine_inner.geom,
                  title="Final Inner Geometry")

        plot_geom(False,  # for developer
                  p, machine_outer.geom,
                  title="Final Outer Geometry")

        if show_plots:
            p.render_elements(machine_inner.geom, Shape,
                              draw_inside=True, title=inner_title,
                              rows=3, cols=2, num=5, show=False,
                              with_corners=False,
                              with_nodes=False,
                              neighbors=False,
                              write_id=write_id,
                              draw_phi=True,
                              fill_areas=True)

            p.render_elements(machine_outer.geom, Shape,
                              draw_inside=True, title=outer_title,
                              rows=3, cols=2, num=6, show=False,
                              with_corners=False,
                              with_nodes=False,
                              neighbors=False,
                              write_id=write_id,
                              draw_phi=True,
                              fill_areas=True)
        elif small_plots:
            #p.figure(figsize=(9, 5)).suptitle(input_file.name, fontsize=16)
            p.render_elements(machine_inner.geom, Shape,
                              draw_inside=True, title=inner_title,
                              rows=1, cols=2, num=1, show=False,
                              with_corners=False,
                              with_nodes=False,
                              neighbors=False,
                              write_id=write_id,
                              draw_phi=True,
                              fill_areas=True)

            p.render_elements(machine_outer.geom, Shape,
                              draw_inside=True, title=outer_title,
                              rows=1, cols=2, num=2, show=False,
                              with_corners=False,
                              with_nodes=False,
                              neighbors=False,
                              write_id=write_id,
                              draw_phi=True,
                              fill_areas=True)

        if show_plots or small_plots:
            if write_png:
                p.write_plot(basename)
            else:
                p.show_plot()

        if show_areas:
            p.render_elements(machine_inner.geom, Shape,
                              title=inner_name,
                              show=True,
                              draw_inside=True,
                              neighbors=True,
                              fill_areas=True)
            p.render_areas(machine_inner.geom,
                           title=inner_name,
                           with_nodes=True,
                           single_view=True)

            p.render_elements(machine_outer.geom, Shape,
                              title=outer_name,
                              show=True,
                              draw_inside=True,
                              neighbors=True,
                              fill_areas=True)
            p.render_areas(machine_outer.geom,
                           title=outer_name,
                           with_nodes=True,
                           single_view=True)

        params = create_femag_parameters(machine_inner,
                                         machine_outer,
                                         nodedist)

        if write_fsl:
            if machine_inner.is_full() or machine_outer.is_full():
                logger.warning("it's not possible to create fsl-file")
                return None

            mtype = 'EESM' if EESM else 'PMSM'
            fslrenderer = FslRenderer(basename, mtype)
            inner = fslrenderer.render(machine_inner, inner=True)
            outer = fslrenderer.render(machine_outer, outer=True)
            if write_fsl_single:
                inner_single = fslrenderer.render(machine_inner, inner=True,
                                                  standalone=True)
                outer_single = fslrenderer.render(machine_outer, outer=True,
                                                  standalone=True)
            else:
                inner_single = None
                outer_single = None
            if full_model:
                params['num_sl_gen'] = params.get('tot_num_slot', 0)
            params['agndst'] = agndst(params.get('da1'),
                                      params.get('da2'),
                                      params.get('tot_num_slot'),
                                      params.get('num_poles'),
                                      params.get('nodedist'))

            if params['external_rotor']:
                conv['fsl_rotor'] = outer
                if outer_single:
                    conv['fsl_rotor_single'] = outer_single
                conv['fsl_stator'] = inner
                if inner_single:
                    conv['fsl_stator_single'] = inner_single
            else:
                conv['fsl_rotor'] = inner
                if inner_single:
                    conv['fsl_rotor_single'] = inner_single
                conv['fsl_stator'] = outer
                if outer_single:
                    conv['fsl_stator_single'] = outer_single

            conv['fsl'] = fslrenderer.render_main(
                machine_inner, machine_outer,
                inner, outer,
                params)
    else:
        # No airgap found. This must be an inner or outer part
        logger.info("=== no airgap found ===")

        name = "No_Airgap"
        inner = False
        outer = False
        params = None

        if part:
            if part[1] == 'in':
                name = inner_name
                inner = True
            else:
                name = outer_name
                outer = True

        machine = symmetry_search(machine,
                                  plt=p,  # plot
                                  kind=name,
                                  is_inner=inner,
                                  is_outer=outer,
                                  mindist=mindist,
                                  symtol=symtol,
                                  sympart=sympart,
                                  show_plots=show_plots,
                                  rows=3,  # rows
                                  cols=2,  # cols
                                  num=3)   # start num
        machine.delete_tiny_elements(mindist)

        if da > 0.0 or dy > 0.0:
            if inner:
                r_out = da / 2.0
                r_in = dy / 2.0
            elif outer:
                r_out = dy / 2.0
                r_in = da / 2.0
            else:
                r_out = 0.0
                r_in = 0.0
            if machine.cut_is_possible(r_in, r_out):
                logger.debug("make a cut")
                machine = machine.cut(r_in, r_out)

        if part:
            if part[0] == 'stator':
                machine.geom.set_stator()
                machine.set_inner_or_outer(part[1])
                machine.search_stator_subregions(single=True)
                machine = build_machine_stator(machine,
                                               inner,
                                               mindist,
                                               p,
                                               EESM=EESM,
                                               single=True)
                params = create_femag_parameters_stator(machine,
                                                        part[1])
            else:
                machine.geom.set_rotor()
                machine.set_inner_or_outer(part[1])
                machine.search_rotor_subregions(EESM, single=True)
                machine = build_machine_rotor(machine,
                                              inner,
                                              mindist,
                                              p,
                                              EESM=EESM,
                                              single=True)

                params = create_femag_parameters_rotor(machine,
                                                       part[1])
        else:
            machine.search_subregions(EESM, single=True)

        machine.create_inner_corner_areas()

        logger.info("***** END of work: %s *****", basename)

        if machine.geom.is_rotor():
            title = "Rotor"
        else:
            title = "Stator"

        if show_plots:
            p.render_elements(machine.geom, Shape,
                              draw_inside=True, title=title,
                              rows=3, cols=2, num=5, show=False,
                              with_corners=False,
                              with_nodes=False,
                              neighbors=False,
                              write_id=write_id,
                              draw_phi=True,
                              fill_areas=True)
        elif small_plots:
            #p.figure().suptitle(input_file.name, fontsize=16)
            p.render_elements(machine.geom, Shape,
                              draw_inside=True, title=title,
                              rows=1, cols=1, num=1, show=False,
                              with_corners=False,
                              with_nodes=False,
                              neighbors=False,
                              write_id=write_id,
                              draw_phi=True,
                              fill_areas=True)

        if show_plots or small_plots:
            if write_png:
                p.write_plot(basename)
            else:
                p.show_plot()

        if show_areas:
            p.render_elements(machine.geom, Shape,
                              title=name,
                              show=True,
                              draw_inside=True,
                              neighbors=True,
                              fill_areas=True)
            p.render_areas(machine.geom,
                           title=name,
                           with_nodes=True,
                           single_view=True)

        if write_fsl:
            if machine.is_full():
                logger.warning("it's not possible to create fsl-file")
                return None

            mtype = 'EESM' if EESM else 'PMSM'
            fslrenderer = FslRenderer(basename, mtype)
            conv['fsl'] = fslrenderer.render(machine, inner, outer, standalone=True)

    if params is not None:
        conv.update(params)

    if write_fsl:
        logger.debug("Write fsl")
        if conv and conv['fsl']:
            with io.open(basename + '.fsl', 'w',
                         encoding='utf-8') as f:
                f.write('\n'.join(conv['fsl']))
        else:
            logger.warning("No fsl data available")

        if conv:
            if conv.get('fsl_rotor_single', None):
                with io.open(basename + '_ROTOR.fsl', 'w',
                             encoding='utf-8') as f:
                    f.write('\n'.join(conv['fsl_rotor_single']))
                del conv['fsl_rotor_single']
            if conv.get('fsl_stator_single', None):
                with io.open(basename + '_STATOR.fsl', 'w',
                             encoding='utf-8') as f:
                    f.write('\n'.join(conv['fsl_stator_single']))
                del conv['fsl_stator_single']

    conv['name'] = basename
    t = timer.stop("-- all done in %0.4f seconds --", info=True)
    journal.put('time_total', t)
    journal.set('success', True)
    journal.write_journal()
    return conv


def _create_rotor_parameters(machine):
    rotor = {
        'min_radius': machine.geom.min_radius,
        'max_radius': machine.geom.max_radius,
        'mags': machine.geom.magnets_minmax_list(),
        'fd_wnds': machine.geom.fd_windings_minmax_list()
    }
    shaft_min, shaft_max = machine.geom.shaft_minmax()
    if shaft_max > 0.0:
        rotor['shaft_min'] = shaft_min
        rotor['shaft_max'] = shaft_max
    if shaft_max > rotor['min_radius']:
        rotor['min_radius'] = shaft_max
    return rotor


def _create_stator_parameters(machine):
    stator = {
        'min_radius': machine.geom.min_radius,
        'max_radius': machine.geom.max_radius,
        'wnds': machine.geom.windings_minmax_list()
    }
    shaft_min, shaft_max = machine.geom.shaft_minmax()
    if shaft_max > 0.0:
        stator['shaft_min'] = shaft_min
        stator['shaft_max'] = shaft_max
    if shaft_max > stator['min_radius']:
        stator['min_radius'] = shaft_max
    return stator


def create_femag_parameters(m_inner, m_outer, nodedist=1):
    if not (m_inner and m_outer):
        logger.warning("inner %s outer %s", m_inner, m_outer)
        return {}

    params = {}
    geom_inner = m_inner.geom
    geom_outer = m_outer.geom

    parts_inner = int(m_inner.get_symmetry_part())
    parts_outer = int(m_outer.get_symmetry_part())

    if m_inner.geom.is_rotor():
        geom_slots = geom_outer
        geom_poles = geom_inner
        num_slots = int(m_outer.get_symmetry_part())
        num_poles = int(m_inner.get_symmetry_part())
    else:
        geom_slots = geom_inner
        geom_poles = geom_outer
        num_slots = int(m_inner.get_symmetry_part())
        num_poles = int(m_outer.get_symmetry_part())

    slot_area = 0
    slot_area = geom_slots.area_size_of_type(TYPE_WINDINGS)
    num_sl_gen = int(geom_slots.get_symmetry_copies()+1)
    alfa_slot = geom_slots.get_alfa()
    alfa_pole = geom_poles.get_alfa()

    params['tot_num_slot'] = num_slots
    params['slot_area'] = slot_area
    params['num_sl_gen'] = num_sl_gen
    params['num_poles'] = num_poles
    params['nodedist'] = nodedist
    params['external_rotor'] = m_outer.geom.is_rotor()
    params['dy1'] = 2*geom_outer.max_radius
    params['da1'] = 2*geom_outer.min_radius
    params['da2'] = 2*geom_inner.max_radius
    params['dy2'] = 2*geom_inner.min_radius
    params['alfa_slot'] = alfa_slot
    params['alfa_pole'] = alfa_pole

    if m_inner.geom.is_rotor():
        params['rotor'] = _create_rotor_parameters(m_inner)
        params['stator'] = _create_stator_parameters(m_outer)
    else:
        params['rotor'] = _create_rotor_parameters(m_outer)
        params['stator'] = _create_stator_parameters(m_inner)

    if num_slots == 0 or num_poles == 0:
        if num_slots == 0:
            logger.warning("No slots found")
        if num_poles == 0:
            logger.warning("No poles found")
        logger.warning("Model not ready for femag")
        return {'error': "Model not ready for femag"}

    if not np.isclose(alfa_slot * num_slots,
                      alfa_pole * num_poles):
        logger.warning("slots and poles dont match")
        return {'error': "Model not ready for femag"}

    return params


def create_femag_parameters_stator(motor, position):
    params = {}
    num_slots = motor.get_num_slots()
    params['tot_num_slot'] = int(num_slots)
    if position == 'in':
        params['da2'] = 2*motor.geom.max_radius
        params['dy2'] = 2*motor.geom.min_radius
    else:
        params['dy1'] = 2*motor.geom.max_radius
        params['da1'] = 2*motor.geom.min_radius
    params['slot_area'] = motor.slot_area()
    params['stator'] = _create_stator_parameters(motor)
    #params['machine'] = motor
    return params


def create_femag_parameters_rotor(motor, position):
    params = {}
    num_poles = motor.get_num_poles()
    params['num_poles'] = int(num_poles)
    if position == 'in':
        params['da2'] = 2*motor.geom.max_radius
        params['dy2'] = 2*motor.geom.min_radius
    else:
        params['dy1'] = 2*motor.geom.max_radius
        params['da1'] = 2*motor.geom.min_radius
    params['slot_area'] = motor.slot_area()
    params['rotor'] = _create_rotor_parameters(motor)
    #params['machine'] = motor
    return params
